# -*- coding: utf-8 -*-
"""

    mslib.msui.satellite_dockwidget
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Control to load satellite track predictions into the top view.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
import os
from datetime import datetime, timedelta
import numpy as np

from mslib.msui.mss_qt import get_open_filename
from mslib.msui.mss_qt import ui_satellite_dockwidget as ui
from PyQt5 import QtWidgets
from mslib.utils.config import save_settings_qsettings, load_settings_qsettings
from fs import open_fs


def read_nasa_satellite_prediction(fname):
    """Read a text file as downloaded from the NASA satellite prediction tool.

    This method reads satellite overpass predictions in ASCII format as
    downloaded from http://www-air.larc.nasa.gov/tools/predict.htm.

    Returns a list of dictionaries with keys
      -- utc: Nx1 array with utc times as datetime objects
      -- satpos: Nx2 array with lon/lat (x/y) of satellite positions
      -- heading: Nx1 array with satellite headings in degrees
      -- swath_left: Nx2 array with lon/lat of left swath boundary
      -- swath_right: Nx2 array with lon/lat of right swath boundary
    Each dictionary represents a separate overpass.

    All arrays are masked arrays, note that missing values are common. Filter
    out missing values with numpy.ma.compress_rows().

    NOTE: ****** LON in the 'predict' files seems to be wrong --> needs to be
                 multiplied by -1. ******
    """
    # Read the file into a list of strings.
    _dirname, _name = os.path.split(fname)
    _fs = open_fs(_dirname)
    satfile = _fs.open(_name, 'r')
    satlines = satfile.readlines()
    satfile.close()

    # Determine the date from the first line.
    date = datetime.strptime(satlines[0].split()[0], "%Y/%m/%d")
    basedate = datetime.strptime("", "")

    # "result" will store the individual overpass segments.
    result = []
    segment = {"utc": [], "satpos": [], "heading": [],
               "swath_left": [], "swath_right": []}

    # Define a time difference that specifies when to start a new segment.
    # If the time between to subsequent points in the file is larger than
    # this time, a new segment will be started.
    seg_diff_time = timedelta(minutes=10)

    # Loop over data lines. Either append point to current segment or start
    # new segment. Before storing segments to the "result" list, convert
    # to masked arrays.
    for line in satlines[2:]:
        values = line.split()
        time = date + (datetime.strptime(values[0], "%H:%M:%S") - basedate)

        if len(segment["utc"]) == 0 or (time - segment["utc"][-1]) < seg_diff_time:
            segment["utc"].append(time)
            segment["satpos"].append([-1. * float(values[2]), float(values[1])])
            segment["heading"].append(float(values[3]))
            if len(values) == 8:
                segment["swath_left"].append([-1. * float(values[5]), float(values[4])])
                segment["swath_right"].append([-1. * float(values[7]), float(values[6])])
            else:
                # TODO 20100504: workaround for instruments without swath
                segment["swath_left"].append([-1. * float(values[2]), float(values[1])])
                segment["swath_right"].append([-1. * float(values[2]), float(values[1])])

        else:
            segment["utc"] = np.array(segment["utc"])
            segment["satpos"] = np.ma.masked_equal(segment["satpos"], -999.)
            segment["heading"] = np.ma.masked_equal(segment["heading"], -999.)
            segment["swath_left"] = np.ma.masked_equal(segment["swath_left"], -999.)
            segment["swath_right"] = np.ma.masked_equal(segment["swath_right"], -999.)
            result.append(segment)
            segment = {"utc": [], "satpos": [], "heading": [],
                       "swath_left": [], "swath_right": []}

    return result


class SatelliteControlWidget(QtWidgets.QWidget, ui.Ui_SatelliteDockWidget):
    def __init__(self, parent=None, view=None):
        super(SatelliteControlWidget, self).__init__(parent)
        self.setupUi(self)
        self.view = view

        # Currently loaded satellite overpass segments.
        self.overpass_segments = None

        # Connect slots and signals.
        self.btSelectFile.clicked.connect(self.select_file)
        self.btLoadFile.clicked.connect(self.load_file)
        self.cbSatelliteOverpasses.currentIndexChanged.connect(self.plot_overpass_track)

        self.settings_tag = "satellitedock"
        settings = load_settings_qsettings(self.settings_tag, {"filename": ""})
        self.leFile.setText(settings["filename"])

    def select_file(self):
        """
        Slot that opens a file dialog to choose a file with satellite
        overpass predictions.
        """
        filename = get_open_filename(
            self, "Open NASA satellite overpass prediction",
            os.path.join(os.path.dirname(self.leFile.text())), "All Files (*)",
            pickertag="filepicker_default")
        if not filename:
            return
        self.leFile.setText(filename)
        save_settings_qsettings(self.settings_tag, {"filename": filename})

    def load_file(self):
        """
        Load the file specified in leFile and fill the combobox with the
        available track segments.
        """
        filename = self.leFile.text()
        logging.debug("loading satellite overpasses in file '%s'", filename)

        try:
            overpass_segments = read_nasa_satellite_prediction(filename)
        except (IOError, OSError, ValueError) as ex:
            logging.error("Problem accessing '%s' file", filename)
            QtWidgets.QMessageBox.critical(self, self.tr("Satellite Overpass Tool"),
                                           self.tr(f"ERROR:\n{type(ex)}\n{ex}"))
        else:
            logging.debug("read %i segments", len(overpass_segments))

            self.cbSatelliteOverpasses.clear()
            items = [f"{seg['utc'][0]} to {seg['utc'][-1]}"
                     for seg in overpass_segments]
            items.insert(0, "None (select item to plot)")
            items.insert(1, "All tracks")
            self.cbSatelliteOverpasses.addItems(items)

            self.overpass_segments = overpass_segments

    def plot_overpass_track(self, index):
        """
        """
        index -= 2
        logging.debug("plotting satellite overpass #%i", index)
        segments = []
        if 0 <= index < len(self.overpass_segments):
            segments = [self.overpass_segments[index]]
        elif index == -1:
            segments = self.overpass_segments
        if self.view is not None:
            self.view.plot_satellite_overpass(segments)



# -*- coding: utf-8 -*-
"""

    mslib.utils._tests.test_config
    ~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to test mslib.utils.config

    This file is part of mss.

    :copyright: Copyright 2016-2017 Reimar Bauer
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
import mslib.utils.config as config
import os
import fs
import pytest

from mslib import utils
from mslib.utils.config import MissionSupportSystemDefaultConfig as mss_default
from mslib.utils.config import config_loader, read_config_file
from mslib._tests.constants import MSS_CONFIG_PATH
from mslib._tests.utils import create_mss_settings_file

LOGGER = logging.getLogger(__name__)


class TestSettingsSave(object):
    """
    tests save_settings_qsettings and load_settings_qsettings from ./utils.py
    # TODO make sure do a clean setup, not inside the 'mss' config file.
    """
    tag = "test_automated"

    def test_save_settings(self):
        settings = {'foo': 'bar'}
        config.save_settings_qsettings(self.tag, settings, ignore_test=True)

    def test_load_settings(self):
        settings = {'foo': 'bar'}
        config.save_settings_qsettings(self.tag, settings, ignore_test=True)
        settings = config.load_settings_qsettings(self.tag, ignore_test=True)
        assert isinstance(settings, dict)
        assert settings["foo"] == "bar"


class TestConfigLoader(object):
    """
    tests config file for client
    """

    def teardown(self):
        if fs.open_fs(MSS_CONFIG_PATH).exists("mss_settings.json"):
            fs.open_fs(MSS_CONFIG_PATH).remove("mss_settings.json")

    def test_option_types(self):
        # check if all config options are added to the appropriate type of options
        config_keys = set(config_loader(default=True).keys())
        option_types = set(mss_default.fixed_dict_options +
                           mss_default.key_value_options +
                           list(mss_default.dict_option_structure.keys()) +
                           list(mss_default.list_option_structure.keys()))
        assert config_keys == option_types

    def test_default_config(self):
        data = config_loader(default=True)
        assert isinstance(data, dict)
        assert data == config_loader(default=True)
        assert data["num_labels"] == 10
        assert data["num_interpolation_points"] == 201

    def test_default_config_dataset(self):
        default_data = config_loader(default=True)
        num_labels = config_loader(dataset="num_labels", default=True)
        assert num_labels == default_data["num_labels"]

    def test_default_config_wrong_file(self):
        # return default if no access to config file given
        with pytest.raises(FileNotFoundError):
            read_config_file(path="foo.json")

    def test_sample_config_file(self):
        utils_path = os.path.dirname(os.path.abspath(utils.__file__))
        config_file = os.path.join(
            utils_path,
            '../',
            '../',
            'docs',
            'samples',
            'config',
            'mss',
            'mss_settings.json.sample',
        )
        read_config_file(path=config_file)
        data = config_loader(dataset="new_flighttrack_flightlevel")
        assert data == 250
        with pytest.raises(KeyError):
            config_loader(dataset="UNDEFINED")
        with pytest.raises(KeyError):
            assert config_loader(dataset="UNDEFINED")
        with pytest.raises(FileNotFoundError):
            config_file = os.path.join(
                utils_path,
                '../',
                '../',
                'docs',
                'samples',
                'config',
                'mss',
                'non_existent_mss_settings.json.sample',
            )
            read_config_file(config_file)

    def test_existing_empty_config_file(self):
        """
        on a user defined empty mss_settings_json this test should return the default value for num_labels
        """
        create_mss_settings_file('{ }')
        if not fs.open_fs(MSS_CONFIG_PATH).exists("mss_settings.json"):
            pytest.skip('undefined test mss_settings.json')
        with fs.open_fs(MSS_CONFIG_PATH) as file_dir:
            file_content = file_dir.readtext("mss_settings.json")
        assert ":" not in file_content
        default_data = config_loader(default=True)
        config_file = fs.path.combine(MSS_CONFIG_PATH, "mss_settings.json")
        read_config_file(path=config_file)
        data = config_loader()
        assert data["num_labels"] == default_data["num_labels"]
        num_labels = config_loader(dataset="num_labels")
        assert num_labels == default_data["num_labels"]
        with pytest.raises(KeyError):
            config_loader(dataset="UNDEFINED")
        with pytest.raises(KeyError):
            assert config_loader(dataset="UNDEFINED")

    def test_existing_config_file_different_parameters(self):
        """
        on a user defined mss_settings_json without a defined num_labels this test should return its default value
        """
        create_mss_settings_file('{"num_interpolation_points": 20 }')
        if not fs.open_fs(MSS_CONFIG_PATH).exists("mss_settings.json"):
            pytest.skip('undefined test mss_settings.json')
        with fs.open_fs(MSS_CONFIG_PATH) as file_dir:
            file_content = file_dir.readtext("mss_settings.json")
        assert "num_labels" not in file_content
        default_data = config_loader(default=True)
        config_file = fs.path.combine(MSS_CONFIG_PATH, "mss_settings.json")
        read_config_file(path=config_file)
        data = config_loader()
        assert data["num_labels"] == default_data["num_labels"]
        num_labels = config_loader(dataset="num_labels")
        assert num_labels == default_data["num_labels"]
        num_interpolation_points = config_loader(dataset="num_interpolation_points")
        assert num_interpolation_points == 20
        assert data["num_interpolation_points"] == 20
        with pytest.raises(KeyError):
            config_loader(dataset="UNDEFINED")
        with pytest.raises(KeyError):
            assert config_loader(dataset="UNDEFINED")

    def test_existing_config_file_defined_parameters(self):
        """
        on a user defined mss_settings_json without a defined num_labels this test should return its default value
        """
        create_mss_settings_file('{"num_interpolation_points": 201, "num_labels": 10 }')
        if not fs.open_fs(MSS_CONFIG_PATH).exists("mss_settings.json"):
            pytest.skip('undefined test mss_settings.json')
        with fs.open_fs(MSS_CONFIG_PATH) as file_dir:
            file_content = file_dir.readtext("mss_settings.json")
        assert "num_labels" in file_content
        config_file = fs.path.combine(MSS_CONFIG_PATH, "mss_settings.json")
        read_config_file(path=config_file)
        num_labels = config_loader(dataset="num_labels")
        assert num_labels == 10
        with pytest.raises(KeyError):
            config_loader(dataset="UNDEFINED")
        with pytest.raises(KeyError):
            assert config_loader(dataset="UNDEFINED")

    def test_existing_config_file_invalid_parameters(self):
        """
        on a user defined mss_settings_json with duplicate and empty keys should raise FatalUserError
        """
        create_mss_settings_file('{"num_interpolation_points": 201, "num_interpolation_points": 10 }')
        if not fs.open_fs(MSS_CONFIG_PATH).exists("mss_settings.json"):
            pytest.skip('undefined test mss_settings.json')
        with fs.open_fs(MSS_CONFIG_PATH) as file_dir:
            file_content = file_dir.readtext("mss_settings.json")
        assert "num_interpolation_points" in file_content
        config_file = fs.path.combine(MSS_CONFIG_PATH, "mss_settings.json")
        with pytest.raises(utils.FatalUserError):
            read_config_file(path=config_file)

        create_mss_settings_file('{"": 201, "num_labels": 10 }')
        if not fs.open_fs(MSS_CONFIG_PATH).exists("mss_settings.json"):
            pytest.skip('undefined test mss_settings.json')
        with fs.open_fs(MSS_CONFIG_PATH) as file_dir:
            file_content = file_dir.readtext("mss_settings.json")
        assert "num_labels" in file_content
        with pytest.raises(utils.FatalUserError):
            read_config_file(path=config_file)



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_wms_capabilities
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests msui.wms_capabilities

    This file is part of mss.

    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2017-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
import mock

from PyQt5 import QtWidgets, QtTest, QtCore
import mslib.msui.wms_capabilities as wc


class Test_WMSCapabilities(object):

    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)
        self.capabilities = mock.Mock()
        self.capabilities.capabilities_document = u"Hölla die Waldfee".encode("utf-8")
        self.capabilities.provider = mock.Mock()
        self.capabilities.identification = mock.Mock()
        self.capabilities.provider.contact = mock.Mock()
        self.capabilities.provider.contact.name = None
        self.capabilities.provider.contact.organization = None
        self.capabilities.provider.contact.email = None
        self.capabilities.provider.contact.address = None
        self.capabilities.provider.contact.postcode = None
        self.capabilities.provider.contact.city = None

    def start_window(self):
        self.window = wc.WMSCapabilitiesBrowser(
            url="http://example.com",
            capabilities=self.capabilities)
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)

    def teardown(self):
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_window_start(self, mockbox):
        self.start_window()
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_window_contact_none(self, mockbox):
        self.capabilities.provider.contact = None
        self.start_window()
        assert mockbox.critical.call_count == 0

    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_switch_view(self, mockbox):
        self.start_window()
        QtTest.QTest.mouseClick(self.window.cbFullView, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0
        QtTest.QTest.mouseClick(self.window.cbFullView, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 0



# -*- coding: utf-8 -*-
"""

    mslib.mswms._tests.test_dataaccess
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests mswms.dataaccess

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2017 Reimar Bauer
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
from datetime import datetime

import mock

from mslib.mswms.dataaccess import DefaultDataAccess, CachedDataAccess
from mslib._tests.constants import DATA_DIR


class Test_DefaultDataAccess(object):
    def setup(self):
        self.dut = DefaultDataAccess(DATA_DIR, "EUR_LL015")
        self.dut.setup()

    def test_get_filename(self):
        filename = self.dut.get_filename("air_pressure", "ml",
                                         datetime(2012, 10, 17, 12, 0),
                                         datetime(2012, 10, 17, 18, 0))
        assert filename == "20121017_12_ecmwf_forecast.P_derived.EUR_LL015.036.ml.nc"

        filename = self.dut.get_filename("air_pressure", "ml",
                                         datetime(2012, 10, 17, 12, 0),
                                         datetime(2012, 10, 17, 18, 0),
                                         fullpath=True)
        assert filename == os.path.join(DATA_DIR, filename)

    def test_get_datapath(self):
        assert self.dut.get_datapath() == DATA_DIR

    def test_get_all_datafiles(self):
        all_files = self.dut.get_all_datafiles()
        assert sorted(all_files) == sorted(os.listdir(DATA_DIR))

    def test_get_init_times(self):
        all_init_times = self.dut.get_init_times()
        assert all_init_times == [datetime(2012, 10, 17, 12, 0)]

    def test_mfDatasetArgs(self):
        mfDatasetArgs = self.dut.mfDatasetArgs()
        assert mfDatasetArgs == {'skip_dim_check': []}
        mfDatasetArgs2 = DefaultDataAccess(DATA_DIR, "EUR_LL015", skip_dim_check=["time1"]).mfDatasetArgs()
        assert mfDatasetArgs2 == {'skip_dim_check': ['time1']}

    def test_get_valid_times(self):
        valid_times = self.dut.get_valid_times("air_pressure", "ml", datetime(2012, 10, 17, 12, 0))
        assert valid_times == [datetime(2012, 10, 17, 12, 0),
                               datetime(2012, 10, 17, 18, 0),
                               datetime(2012, 10, 18, 0, 0),
                               datetime(2012, 10, 18, 6, 0),
                               datetime(2012, 10, 18, 12, 0),
                               datetime(2012, 10, 18, 18, 0),
                               datetime(2012, 10, 19, 0, 0)]

    def test_get_all_valid_times(self):
        all_valid_times = self.dut.get_all_valid_times("air_pressure", "ml")
        assert sorted(all_valid_times) == \
            sorted([datetime(2012, 10, 18, 18, 0),
                    datetime(2012, 10, 18, 0, 0),
                    datetime(2012, 10, 17, 12, 0),
                    datetime(2012, 10, 18, 6, 0),
                    datetime(2012, 10, 17, 18, 0),
                    datetime(2012, 10, 18, 12, 0),
                    datetime(2012, 10, 19, 0, 0)])


class Test_CachedDataAccess(Test_DefaultDataAccess):
    """
    Reuse default testcases and add some more
    """

    def setup(self):
        self.dut = CachedDataAccess(DATA_DIR, "EUR_LL015")
        self.dut.setup()

    def test_cache_full(self):
        self.dut._parse_file = mock.MagicMock()
        self.dut._add_to_filetree = mock.MagicMock()
        self.dut.setup()
        assert self.dut._parse_file.call_count == 0

    def test_cache_modified(self):
        self.dut._parse_file = mock.MagicMock()
        self.dut._add_to_filetree = mock.MagicMock()
        n = len(self.dut.get_all_datafiles())
        fn = list(self.dut._file_cache.keys())[0]
        self.dut._file_cache[fn] = (
            self.dut._file_cache[fn][0] + 1,
            self.dut._file_cache[fn][1])
        self.dut.setup()
        self.dut._parse_file.assert_called_once_with(fn)
        assert self.dut._add_to_filetree.call_count == n

    def test_cache_incomplete(self):
        self.dut._parse_file = mock.MagicMock()
        self.dut._add_to_filetree = mock.MagicMock()
        n = len(self.dut.get_all_datafiles())
        fn = list(self.dut._file_cache.keys())[0]
        del self.dut._file_cache[fn]
        self.dut.setup()
        assert self.dut._parse_file.call_count == 1
        assert self.dut._add_to_filetree.call_count == n

    def test_cache_too_large(self):
        self.dut._file_cache["nothere"] = [0, {}]
        self.dut.setup()
        assert "nothere" not in self.dut._file_cache


class Test_DefaultDataAccessNoInit(object):
    def setup(self):
        self.dut = DefaultDataAccess(DATA_DIR, "EUR_LL015", uses_init_time=False)
        self.dut.setup()

    def test_get_init_times(self):
        all_init_times = self.dut.get_init_times()
        assert all_init_times == [None]



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_wms_multilayers.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MultilayersDialog(object):
    def setupUi(self, MultilayersDialog):
        MultilayersDialog.setObjectName("MultilayersDialog")
        MultilayersDialog.resize(892, 396)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(MultilayersDialog)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.cbMultilayering = QtWidgets.QCheckBox(MultilayersDialog)
        self.cbMultilayering.setObjectName("cbMultilayering")
        self.horizontalLayout_11.addWidget(self.cbMultilayering)
        self.line = QtWidgets.QFrame(MultilayersDialog)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_11.addWidget(self.line)
        self.lFilter = QtWidgets.QLabel(MultilayersDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lFilter.sizePolicy().hasHeightForWidth())
        self.lFilter.setSizePolicy(sizePolicy)
        self.lFilter.setObjectName("lFilter")
        self.horizontalLayout_11.addWidget(self.lFilter)
        self.leMultiFilter = QtWidgets.QLineEdit(MultilayersDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.leMultiFilter.sizePolicy().hasHeightForWidth())
        self.leMultiFilter.setSizePolicy(sizePolicy)
        self.leMultiFilter.setMaximumSize(QtCore.QSize(16777215, 20))
        self.leMultiFilter.setObjectName("leMultiFilter")
        self.horizontalLayout_11.addWidget(self.leMultiFilter)
        self.verticalLayout_3.addLayout(self.horizontalLayout_11)
        self.listLayers = QtWidgets.QTreeWidget(MultilayersDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.listLayers.sizePolicy().hasHeightForWidth())
        self.listLayers.setSizePolicy(sizePolicy)
        self.listLayers.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.listLayers.setAutoFillBackground(False)
        self.listLayers.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.listLayers.setDragEnabled(False)
        self.listLayers.setDragDropOverwriteMode(False)
        self.listLayers.setAlternatingRowColors(True)
        self.listLayers.setAnimated(False)
        self.listLayers.setHeaderHidden(False)
        self.listLayers.setColumnCount(4)
        self.listLayers.setObjectName("listLayers")
        self.listLayers.headerItem().setText(0, "Name")
        self.listLayers.header().setVisible(True)
        self.listLayers.header().setCascadingSectionResizes(True)
        self.listLayers.header().setStretchLastSection(False)
        self.verticalLayout_3.addWidget(self.listLayers)
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.label_7 = QtWidgets.QLabel(MultilayersDialog)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_12.addWidget(self.label_7)
        self.cbWMS_URL = QtWidgets.QComboBox(MultilayersDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbWMS_URL.sizePolicy().hasHeightForWidth())
        self.cbWMS_URL.setSizePolicy(sizePolicy)
        self.cbWMS_URL.setStyleSheet("QComboBox QAbstractItemView { min-width: 800px; }")
        self.cbWMS_URL.setEditable(True)
        self.cbWMS_URL.setObjectName("cbWMS_URL")
        self.cbWMS_URL.addItem("")
        self.cbWMS_URL.addItem("")
        self.horizontalLayout_12.addWidget(self.cbWMS_URL)
        self.btGetCapabilities = QtWidgets.QPushButton(MultilayersDialog)
        self.btGetCapabilities.setObjectName("btGetCapabilities")
        self.horizontalLayout_12.addWidget(self.btGetCapabilities)
        self.pbViewCapabilities = QtWidgets.QPushButton(MultilayersDialog)
        self.pbViewCapabilities.setObjectName("pbViewCapabilities")
        self.horizontalLayout_12.addWidget(self.pbViewCapabilities)
        self.verticalLayout_3.addLayout(self.horizontalLayout_12)
        self.verticalLayout_4.addLayout(self.verticalLayout_3)
        self.actionCloseWindow = QtWidgets.QAction(MultilayersDialog)
        self.actionCloseWindow.setObjectName("actionCloseWindow")
        MultilayersDialog.addAction(self.actionCloseWindow)

        self.retranslateUi(MultilayersDialog)
        self.actionCloseWindow.triggered.connect(MultilayersDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(MultilayersDialog)

    def retranslateUi(self, MultilayersDialog):
        _translate = QtCore.QCoreApplication.translate
        MultilayersDialog.setWindowTitle(_translate("MultilayersDialog", "Layer List"))
        self.cbMultilayering.setText(_translate("MultilayersDialog", "Multilayering"))
        self.lFilter.setText(_translate("MultilayersDialog", "Layer Filter:"))
        self.listLayers.headerItem().setText(1, _translate("MultilayersDialog", "Style"))
        self.listLayers.headerItem().setText(2, _translate("MultilayersDialog", "Priority"))
        self.listLayers.headerItem().setText(3, _translate("MultilayersDialog", "Colour"))
        self.label_7.setText(_translate("MultilayersDialog", "WMS URL:"))
        self.cbWMS_URL.setToolTip(_translate("MultilayersDialog", "Enter a valid WMS URL here."))
        self.cbWMS_URL.setItemText(0, _translate("MultilayersDialog", "http://localhost:8002/fc_wms"))
        self.cbWMS_URL.setItemText(1, _translate("MultilayersDialog", "http://osm.omniscale.net/proxy/service"))
        self.btGetCapabilities.setToolTip(_translate("MultilayersDialog", "Request the capabilities from the WMS server."))
        self.btGetCapabilities.setText(_translate("MultilayersDialog", "get capabilities"))
        self.pbViewCapabilities.setToolTip(_translate("MultilayersDialog", "Show information on the selected WMS server."))
        self.pbViewCapabilities.setText(_translate("MultilayersDialog", "view"))
        self.actionCloseWindow.setText(_translate("MultilayersDialog", "Close Window"))
        self.actionCloseWindow.setShortcut(_translate("MultilayersDialog", "Ctrl+W"))



# -*- coding: utf-8 -*-
"""

    mslib.mscolab
    ~~~~~~~~~~~~~~

    init module of mscolab

    This file is part of mss.

    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""



# -*- coding: utf-8 -*-
"""

    mslib.mslib.utils
    ~~~~~~~~~~~~~~~~~

    This module provides functions for the wms server

    This file is part of mss.

    :copyright: Copyright 2016 Joern Ungermann
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import numpy as np
import matplotlib
import pint

UR = pint.UnitRegistry()
N_LEVELS = 16


class Targets(object):
    """
    This class defines the names, units, and ranges of supported generic physical
    quantities for vertical and horizontal plots.

    RANGES, UNITS, and THRESHOLDS can be overwritten from outside to determine ranges for plotting in settings.

    Dictionary containing valid value ranges for the different targes. The first level uses CF standard_name as key
    to determine the target. The second level uses either the level type
    ("pl", "ml", ...) as key or "total" for the level-overarching valid range. The level-type has a third level
    using the level altitude as key. The leafs are made of 2-tuples indicating the lowest and the highest valid
    value.
    """
    RANGES = {}

    # List of supported targets using the CF standard_name as unique identifier.
    _TARGETS = [
        "air_temperature",
        "eastward_wind",
        "equivalent_latitude",
        "ertel_potential_vorticity",
        "mean_age_of_air",
        "mole_fraction_of_active_chlorine_in_air",
        "mole_fraction_of_bromine_nitrate_in_air",
        "mole_fraction_of_bromo_methane_in_air",
        "mole_fraction_of_bromochlorodifluoromethane_in_air",
        "mole_fraction_of_bromotrifluoromethane_in_air",
        "mole_fraction_of_carbon_monoxide_in_air",
        "mole_fraction_of_carbon_dioxide_in_air",
        "mole_fraction_of_carbon_tetrachloride_in_air",
        "mole_fraction_of_chlorine_nitrate_in_air",
        "mole_fraction_of_chlorine_dioxide_in_air",
        "mole_fraction_of_cfc11_in_air",
        "mole_fraction_of_cfc113_in_air",
        "mole_fraction_of_cfc12_in_air",
        "mole_fraction_of_ethane_in_air",
        "mole_fraction_of_formaldehyde_in_air",
        "mole_fraction_of_hcfc22_in_air",
        "mole_fraction_of_hydrogen_chloride_in_air",
        "mole_fraction_of_hypobromite_in_air",
        "mole_fraction_of_methane_in_air",
        "mole_fraction_of_nitric_acid_in_air",
        "mole_fraction_of_nitrous_oxide_in_air",
        "mole_fraction_of_nitrogen_dioxide_in_air",
        "mole_fraction_of_nitrogen_monoxide_in_air",
        "mole_fraction_of_ozone_in_air",
        "mole_fraction_of_peroxyacetyl_nitrate_in_air",
        "mole_fraction_of_sulfur_dioxide_in_air",
        "mole_fraction_of_water_vapor_in_air",

        "northward_wind",
        "square_of_brunt_vaisala_frequency_in_air",
        "surface_origin_tracer_from_india_and_china",
        "surface_origin_tracer_from_southeast_asia",
        "surface_origin_tracer_from_east_china",
        "surface_origin_tracer_from_north_india",
        "surface_origin_tracer_from_south_india",
        "surface_origin_tracer_from_north_africa",
        "median_of_age_of_air_spectrum",
        "fraction_below_6months_of_age_of_air_spectrum",
        "fraction_above_24months_of_age_of_air_spectrum",
        "cloud_ice_mixing_ratio",
        "number_concentration_of_ice_crystals_in_air",
        "mean_mass_radius_of_cloud_ice_crystals",
        "maximum_pressure_on_backtrajectory",
    ]

    UNITS = {
        "air_temperature": ("K", 1),
        "eastward_wind": ("1/ms", 1),
        "equivalent_latitude": ("degree N", 1),
        "ertel_potential_vorticity": ("PVU", 1),
        "gravity_wave_temperature_perturbation": ("K", 1),
        "mean_age_of_air": ("month", 1),
        "median_of_age_of_air_spectrum": ("month", 1),
        "northward_wind": ("1/ms", 1),
        "square_of_brunt_vaisala_frequency_in_air": ("1/s²", 1),
        "tropopause_altitude": ("km", 1),
        "cloud_ice_mixing_ratio": ("ppmv", 1),
        "number_concentration_of_ice_crystals_in_air": ("1/cm³", 1),
        "mean_mass_radius_of_cloud_ice_crystals": ("µm", 1),
        "maximum_pressure_on_backtrajectory": ("hPa", 1),
        "maximum_relative_humidity_wrt_ice_on_backtrajectory": ("percent", 1),
    }

    # The THRESHOLDS are used to determine a single colourmap suitable for all plotting purposes (that is vertical
    # and horizontal on all levels. The given thresholds have been manually designed.
    THRESHOLDS = {
        "ertel_potential_vorticity":
            (-1, 0, 1, 2, 4, 6, 9, 12, 15, 25, 40),
        "mole_fraction_of_carbon_monoxide_in_air":
            (10e-9, 20e-9, 30e-9, 40e-9, 50e-9, 60e-9, 70e-9, 80e-9, 90e-9, 100e-9, 300e-9),
        "mole_fraction_of_nitric_acid_in_air":
            (0e-9, 0.3e-9, 0.5e-9, 0.7e-9, 0.9e-9, 1.1e-9, 1.3e-9, 1.5e-9, 2e-9, 4e-9),
        "mole_fraction_of_ozone_in_air":
            (0e-6, 0.02e-6, 0.03e-6, 0.04e-6, 0.06e-6, 0.1e-6, 0.16e-6, 0.25e-6, 0.45e-6, 1e-6, 4e-6),
        "mole_fraction_of_peroxyacetyl_nitrate_in_air":
            (0, 50e-12, 70e-12, 100e-12, 150e-12, 200e-12, 250e-12, 300e-12, 350e-12, 400e-12, 450e-12, 500e-12),
        "mole_fraction_of_water_vapor_in_air":
            (0, 3e-6, 4e-6, 6e-6, 10e-6, 16e-6, 60e-6, 150e-6, 500e-6, 1000e-6),
    }

    for standard_name in _TARGETS:
        if standard_name.startswith("surface_origin_tracer_from_"):
            UNITS[standard_name] = ("percent", 1)

    for standard_name in [
            "mole_fraction_of_carbon_dioxide_in_air",
            "mole_fraction_of_methane_in_air",
            "mole_fraction_of_ozone_in_air",
            "mole_fraction_of_water_vapor_in_air",
    ]:
        UNITS[standard_name] = ("µmol/mol", 1)

    for standard_name in [
            "mole_fraction_of_active_chlorine_in_air",
            "mole_fraction_of_carbon_monoxide_in_air",
            "mole_fraction_of_chlorine_nitrate_in_air",
            "mole_fraction_of_hydrogen_chloride_in_air",
            "mole_fraction_of_nitric_acid_in_air",
            "mole_fraction_of_nitrous_oxide_in_air",
    ]:
        UNITS[standard_name] = ("nmol/mol", 1)

    for standard_name in [
            "mole_fraction_of_bromine_nitrate_in_air",
            "mole_fraction_of_bromo_methane_in_air",
            "mole_fraction_of_bromochlorodifluoromethane_in_air",
            "mole_fraction_of_bromotrifluoromethane_in_air",
            "mole_fraction_of_carbon_tetrachloride_in_air",
            "mole_fraction_of_chlorine_dioxide_in_air",
            "mole_fraction_of_cfc11_in_air",
            "mole_fraction_of_cfc12_in_air",
            "mole_fraction_of_cfc113_in_air",
            "mole_fraction_of_hcfc22_in_air",
            "mole_fraction_of_ethane_in_air",
            "mole_fraction_of_formaldehyde_in_air",
            "mole_fraction_of_hypobromite_in_air",
            "mole_fraction_of_nitrogen_dioxide_in_air",
            "mole_fraction_of_nitrogen_monoxide_in_air",
            "mole_fraction_of_peroxyacetyl_nitrate_in_air",
            "mole_fraction_of_sulfur_dioxide_in_air",
    ]:
        UNITS[standard_name] = ("pmol/mol", 1)

    for standard_name in [
            "fraction_below_6months_of_age_of_air_spectrum",
            "fraction_above_24months_of_age_of_air_spectrum",
    ]:
        UNITS[standard_name] = ("percent", 1)

    TITLES = {
        "ertel_potential_vorticity": "PV",
        "square_of_brunt_vaisala_frequency_in_air": "N²",
        "gravity_wave_temperature_perturbation": "Gravity Wave Temperature Residual",
        "tropopause_altitude": "Thermal Tropopause",
    }
    for standard_name in _TARGETS:
        if standard_name.startswith("mole_fraction_of_") and standard_name.endswith("_in_air"):
            TITLES[standard_name] = standard_name[17:-7].replace("_", " ")
        elif standard_name not in TITLES:
            TITLES[standard_name] = standard_name.replace("_", " ")

    @staticmethod
    def get_targets():
        """
        List to determine what targets are supported.
        Returns:
        list of supported targets
        """
        return Targets._TARGETS

    @staticmethod
    def get_unit(standard_name):
        """
        Returns unit type and scaling factor for target.
        Args:
            standard_name: string of CF standard_name

        Returns:
            Tuple of string describing the unit and scaling factor to apply on data.

        """
        return Targets.UNITS.get(standard_name, (None, 1))

    @staticmethod
    def get_range(standard_name, level="total", typ=None):
        """
        Returns valid range of values for target at given level
        Args:
            standard_name: string of CF standard_name
            level (optional): horizontal level of data
            type (optional): type of data (pl, ml, tl, ...)

        Returns:
            Tuple of lowest and highest valid value
        """
        if standard_name in Targets.RANGES:
            if typ in Targets.RANGES[standard_name]:
                if level in Targets.RANGES[standard_name][typ]:
                    return [_x * Targets.get_unit(standard_name)[1]
                            for _x in Targets.RANGES[standard_name][typ][level]]
                elif level is None:
                    return 0, 0
            if level == "total" and "total" in Targets.RANGES[standard_name]:
                return [_x * Targets.get_unit(standard_name)[1]
                        for _x in Targets.RANGES[standard_name]["total"]]
        if standard_name.startswith("surface_origin_tracer_from_"):
            return 0, 100
        return None, None

    @staticmethod
    def get_thresholds(standard_name, level=None, type=None):
        """
        Returns a list of meaningful values for a BoundaryNorm for plotting.
        Args:
            standard_name: string of CF standard_name
            level (optional): horizontal level of data
            type (optional): type of data (pl, ml, tl, ...)

        Returns:
            Tuple of threshold values to be supplied to a BoundaryNorm.
        """
        try:
            return [_x * Targets.get_unit(standard_name)[1] for _x in Targets.THRESHOLDS[standard_name]]
        except KeyError:
            return None


def get_log_levels(cmin, cmax, levels=N_LEVELS):
    """
    Returns 'levels' levels in a lgarithmic spacing. Takes care of ranges crossing zero and starting/ending at zero.
    Args:
        cmin: minimum value
        cmax: maximum value
        levels (optional): number of levels to be generated

    Returns:
        numpy array of values
    """
    if cmin >= cmax:
        cmin, cmax = cmin - 0.5, cmax + 0.5
    if cmin >= 0:
        if cmin == 0:
            cmin = 0.001 * cmax
        clev = np.exp(np.linspace(np.log(cmin), np.log(cmax), levels))
    elif cmax <= 0:
        if cmax == 0:
            cmax = 0.001 * cmin
        clev = -np.exp(np.linspace(np.log(-cmin), np.log(-cmax), levels))
    else:
        delta = cmax - cmin
        clevlo = -np.exp(
            np.linspace(np.log(-cmin), np.log(max(-cmin, cmax) * 0.001),
                        max(2, 1 + int(levels * -cmin / delta))))
        clevhi = np.exp(np.linspace(np.log(max(-cmin, cmax) * 0.001),
                                    np.log(cmax), max(2, 1 + int(levels * cmax / delta))))
        clev = np.asarray(list(clevlo) + list(clevhi))
    return clev


def get_style_parameters(dataname, style, cmin, cmax, data):
    if cmin is None or cmax is None:
        try:
            cmin, cmax = data.min(), data.max()
        except ValueError:
            cmin, cmax = 0, 1
        if 0 < cmin < 0.05 * cmax:
            cmin = 0.
    cmap = matplotlib.pyplot.cm.rainbow
    ticks = None

    if any(isinstance(_x, np.ma.core.MaskedConstant) for _x in (cmin, cmax)):
        cmin, cmax = 0, 1

    if style == "default":
        clev = np.linspace(cmin, cmax, 16)
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "auto":
        cmin_p = data.min()
        cmax_p = data.max()
        if not any([isinstance(_x, np.ma.core.MaskedConstant) for _x in (cmin_p, cmax_p)]):
            cmin, cmax = cmin_p, cmax_p
        if cmin == cmax:
            cmin, cmax = 0, 1
        if 0 < cmin < 0.05 * cmax:
            cmin = 0.
        clev = np.linspace(cmin, cmax, 16)
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "log":
        clev = get_log_levels(cmin, cmax, 16)
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "autolog":
        cmin_p = data.min()
        cmax_p = data.max()
        if not any([isinstance(_x, np.ma.core.MaskedConstant) for _x in (cmin_p, cmax_p)]):
            cmin, cmax = cmin_p, cmax_p
        if cmin == cmax:
            cmin, cmax = 0, 1
        clev = get_log_levels(cmin, cmax, 16)
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "nonlinear":
        clev = Targets.get_thresholds(dataname)
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style.startswith("ertel_potential_vorticity"):
        colors = [
            (1.0, 0.55000000000000004, 1.0, 1.0),
            (0.82333333333333336, 0.40000000000000002, 1.0, 1.0),
            (0.64666666666666672, 0.25, 1.0, 1.0),
            (0.46999999999999997, 0.10000000000000001, 1.0, 1.0),
            (0.69999999999999996, 1.0, 1.0, 1.0),
            (0.46666666666666667, 0.73333333333333339, 1.0, 1.0),
            (0.23333333333333334, 0.46666666666666667, 1.0, 1.0),
            (0.0, 0.20000000000000001, 1.0, 1.0),
            (0.65000000000000002, 1.0, 0.65000000000000002, 1.0),
            (0.43333333333333335, 0.90000000000000002, 0.43333333333333335, 1.0),
            (0.21666666666666667, 0.80000000000000004, 0.21666666666666667, 1.0),
            (0.0, 0.69999999999999996, 0.0, 1.0),
            (1.0, 1.0, 0.0, 1.0),
            (1.0, 0.375, 0.0, 1.0),
            (0.90000000000000002, 0.0, 0.041666666666666664, 1.0),
            (0.40000000000000002, 0.0, 0.25, 1.0)]
        clev = list(np.arange(0, 4, 0.5)) + list(range(4, 8)) + list(range(8, 18, 2))
        if style[-2:] == "nh":
            cmap = matplotlib.pyplot.cm.colors.ListedColormap(colors, name="pv_map")
            cmap.set_over((0.8, 0.8, 0.8, 1.0))
        else:
            cmap = matplotlib.pyplot.cm.colors.ListedColormap(colors[::-1], name="pv_map")
            cmap.set_under((0.8, 0.8, 0.8, 1.0))
            clev = [-_x for _x in clev[::-1]]
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style.startswith("equivalent_latitude"):
        colors = [
            (1.0, 0.55000000000000004, 1.0, 1.0),
            (0.82333333333333336, 0.40000000000000002, 1.0, 1.0),
            (0.64666666666666672, 0.25, 1.0, 1.0),
            (0.46999999999999997, 0.10000000000000001, 1.0, 1.0),
            (0.69999999999999996, 1.0, 1.0, 1.0),
            (0.46666666666666667, 0.73333333333333339, 1.0, 1.0),
            (0.23333333333333334, 0.46666666666666667, 1.0, 1.0),
            (0.0, 0.20000000000000001, 1.0, 1.0),
            (0.65000000000000002, 1.0, 0.65000000000000002, 1.0),
            (0.43333333333333335, 0.90000000000000002, 0.43333333333333335, 1.0),
            (0.21666666666666667, 0.80000000000000004, 0.21666666666666667, 1.0),
            (0.0, 0.69999999999999996, 0.0, 1.0),
            (1.0, 1.0, 0.0, 1.0),
            (1.0, 0.375, 0.0, 1.0),
            (0.90000000000000002, 0.0, 0.041666666666666664, 1.0),
            (0.40000000000000002, 0.0, 0.25, 1.0)]
        clev = np.arange(5, 86, 5)
        if style[-2:] == "nh":
            cmap = matplotlib.pyplot.cm.colors.ListedColormap(colors)
        else:
            cmap = matplotlib.pyplot.cm.colors.ListedColormap(colors[::-1])
            clev = [-_x for _x in clev[::-1]]
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "gravity_wave_temperature_perturbation":
        cmap = matplotlib.pyplot.cm.Spectral_r
        clev = [-3, -2.5, -2, -1.5, -1, -0.5, 0.5, 1, 1.5, 2, 2.5, 3]
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
        ticks = -3, -2, -1, 1, 2, 3
    elif style == "square_of_brunt_vaisala_frequency_in_air":
        cmap = matplotlib.pyplot.cm.colors.ListedColormap(
            [(1.0, 0.55000000000000004, 1.0, 1.0),
             (0.82333333333333336, 0.40000000000000002, 1.0, 1.0),
             (0.64666666666666672, 0.25, 1.0, 1.0),
             (0.46999999999999997, 0.10000000000000001, 1.0, 1.0),
             (0.69999999999999996, 1.0, 1.0, 1.0),
             (0.0, 0.20000000000000001, 1.0, 1.0),
             (0.65000000000000002, 1.0, 0.65000000000000002, 1.0),
             (0.0, 0.69999999999999996, 0.0, 1.0),
             (1.0, 1.0, 0.0, 1.0),
             (1.0, 0.73529411764705888, 0.0, 1.0),
             (1.0, 0.46323529411764708, 0.0, 1.0),
             (1.0, 0.21568627450980393, 0.0, 1.0),
             (1.0, 0.034313725490196068, 0.0, 1.0),
             (0.8294117647058824, 0.0, 0.071078431372549017, 1.0),
             (0.61176470588235299, 0.0, 0.16176470588235295, 1.0),
             (0.40000000000000002, 0.0, 0.25, 1.0),
             ], name="n2_map")
        cmap.set_over((0.8, 0.8, 0.8, 1.0))
        clev = np.arange(0., 8.5 / 1e4, 0.5 / 1e4)
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "tropopause_altitude":
        cmap = matplotlib.pyplot.cm.terrain
        norm = None
        clev = np.arange(5, 18.1, 0.25)
    elif style == "log_ice_cloud":
        cmap = matplotlib.pyplot.cm.colors.ListedColormap(
            [(0.8, 0.8, 0.8, 1.0),
             (1., 1., 1.0, 1.0),
             (0., 0., 1.0, 1.0),
             (0., 0.949, 0.988, 1.0),
             (0., 1., 0., 1.0),
             (1., 1.0, 0.0, 1.0),
             (1.0, 0., 0.0, 1.0),
             (0.212, 0.114, 0.075, 1.0),
             ], name="ice_map_log")
        if dataname == "number_concentration_of_ice_crystals_in_air":
            clev = np.append(np.array([-1e31, -0.1]),
                             get_log_levels(1e-4, 100., 7))
            ticks = np.append(np.array([0.]), get_log_levels(1e-4, 100., 7))
        if dataname == "cloud_ice_mixing_ratio":
            clev = np.append(np.array([-1e31, -0.1]),
                             get_log_levels(1e-3, 1e3, 7))
            ticks = np.append(np.array([0.]), get_log_levels(1e-3, 1e3, 7))
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    elif style == "ice_cloud":
        cmap = matplotlib.pyplot.cm.colors.ListedColormap(
            [(1., 1., 1.0, 1.0),
             (0., 0., 1.0, 1.0),
             (0., 0.949, 0.988, 1.0),
             (0., 1., 0., 1.0),
             (1., 1.0, 0.0, 1.0),
             (1.0, 0., 0.0, 1.0),
             (0.212, 0.114, 0.075, 1.0),
             ], name="ice_map")
        if dataname == "mean_mass_radius_of_cloud_ice_crystals":
            clev = np.array([0, 1, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100])
        norm = matplotlib.colors.BoundaryNorm(clev, cmap.N)
    else:
        raise RuntimeError(f"Illegal plotting style?! ({style})")
    if clev[0] == clev[-1]:
        cmin, cmax = 0, 1
        clev = np.linspace(0, 1, len(clev))
    return cmin, cmax, clev, cmap, norm, ticks


def get_cbar_label_format(style, maxvalue):
    format = "%.3g"
    if style != "log":
        if 100 <= maxvalue < 10000.:
            format = "%4i"
        elif 10 <= maxvalue < 100.:
            format = "%.1f"
        elif 1 <= maxvalue < 10.:
            format = "%.2f"
        elif 0.1 <= maxvalue < 1.:
            format = "%.3f"
        elif 0.01 <= maxvalue < 0.1:
            format = "%.4f"
    if style == 'log_ice_cloud':
        format = "%.0E"
    return format


def make_cbar_labels_readable(fig, axs):
    """
    Adjust font size of the colorbar labels and put a white background behind them
    such that they are readable in front of any background.
    """
    fontsize = fig.bbox.height * 0.024
    for x in axs.yaxis.majorTicks:
        x.label1.set_path_effects([matplotlib.patheffects.withStroke(linewidth=4, foreground='w')])
        x.label1.set_fontsize(fontsize)



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_mscolab_version_history
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module is used to test mscolab-operation related gui.

    This file is part of mss.

    :copyright: Copyright 2020 Tanish Grover
    :copyright: Copyright 2020-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import sys
import pytest
import mock

from mslib._tests.utils import mscolab_start_server
from mslib.mscolab.conf import mscolab_settings
from PyQt5 import QtCore, QtTest, QtWidgets
from mslib.msui import mscolab
import mslib.msui.mss_pyui as mss_pyui
from mslib.mscolab.mscolab import handle_db_reset
from mslib.mscolab.seed import add_user, get_user, add_operation, add_user_to_operation


PORTS = list(range(20000, 20500))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_MscolabVersionHistory(object):
    def setup(self):
        handle_db_reset()
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        self.userdata = 'UV10@uv10', 'UV10', 'uv10'
        self.operation_name = "europe"
        assert add_user(self.userdata[0], self.userdata[1], self.userdata[2])
        assert add_operation(self.operation_name, "test europe")
        assert add_user_to_operation(path=self.operation_name, emailid=self.userdata[0])
        self.user = get_user(self.userdata[0])
        QtTest.QTest.qWait(500)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = mss_pyui.MSSMainWindow(mscolab_data_dir=mscolab_settings.MSCOLAB_DATA_DIR)
        self.window.show()
        # connect and login to mscolab
        self._connect_to_mscolab()
        self._login(self.userdata[0], self.userdata[2])
        # activate operation and open chat window
        self._activate_operation_at_index(0)
        self.window.actionVersionHistory.trigger()
        QtWidgets.QApplication.processEvents()
        self.version_window = self.window.mscolab.version_window
        assert self.version_window is not None
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()

    def teardown(self):
        if self.window.mscolab.version_window:
            self.window.mscolab.version_window.close()
        if self.window.mscolab.conn:
            self.window.mscolab.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def test_changes(self):
        self._change_version_filter(1)
        len_prev = self.version_window.changes.count()
        # make a changes
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        self.version_window.load_all_changes()
        QtWidgets.QApplication.processEvents()
        len_after = self.version_window.changes.count()
        assert len_prev == (len_after - 2)

    @mock.patch("PyQt5.QtWidgets.QInputDialog.getText", return_value=["MyVersionName", True])
    def test_set_version_name(self, mockbox):
        self._set_version_name()
        QtTest.QTest.qWait(100)
        assert self.version_window.changes.currentItem().version_name == "MyVersionName"
        assert self.version_window.changes.count() == 1

    @mock.patch("PyQt5.QtWidgets.QInputDialog.getText", return_value=["MyVersionName", True])
    def test_version_name_delete(self, mockbox):
        self._set_version_name()
        QtTest.QTest.qWait(100)
        assert self.version_window.changes.currentItem().version_name == "MyVersionName"
        QtTest.QTest.mouseClick(self.version_window.deleteVersionNameBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(500)
        assert self.version_window.changes.count() == 1
        assert self.version_window.changes.currentItem().version_name is None

    @mock.patch("PyQt5.QtWidgets.QMessageBox.question", return_value=QtWidgets.QMessageBox.Yes)
    def test_undo(self, mockbox):
        self._change_version_filter(1)
        # make changes
        for i in range(2):
            self.window.mscolab.waypoints_model.invert_direction()
            QtWidgets.QApplication.processEvents()
            QtTest.QTest.qWait(100)
        self.version_window.load_all_changes()
        QtWidgets.QApplication.processEvents()
        changes_count = self.version_window.changes.count()
        self._activate_change_at_index(1)
        QtTest.QTest.mouseClick(self.version_window.checkoutBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(200)
        new_changes_count = self.version_window.changes.count()
        assert changes_count + 1 == new_changes_count

    def test_refresh(self):
        self._change_version_filter(1)
        changes_count = self.version_window.changes.count()
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        QtTest.QTest.mouseClick(self.version_window.refreshBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        new_changes_count = self.version_window.changes.count()
        assert new_changes_count == changes_count + 2

    def _connect_to_mscolab(self):
        self.connect_window = mscolab.MSColab_ConnectDialog(parent=self.window, mscolab=self.window.mscolab)
        self.window.mscolab.connect_window = self.connect_window
        assert self.connect_window is not None
        self.connect_window.urlCb.setEditText(self.url)
        self.connect_window.show()
        QtTest.QTest.mouseClick(self.connect_window.connectBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(500)

    def _login(self, emailid, password):
        assert self.connect_window is not None
        self.connect_window.loginEmailLe.setText(emailid)
        self.connect_window.loginPasswordLe.setText(password)
        QtTest.QTest.mouseClick(self.connect_window.loginBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(500)

    def _activate_operation_at_index(self, index):
        assert index < self.window.listOperationsMSC.count()
        item = self.window.listOperationsMSC.item(index)
        point = self.window.listOperationsMSC.visualItemRect(item).center()
        QtTest.QTest.mouseClick(self.window.listOperationsMSC.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseDClick(self.window.listOperationsMSC.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()

    def _activate_change_at_index(self, index):
        assert self.version_window is not None
        assert index < self.version_window.changes.count()
        item = self.version_window.changes.item(index)
        point = self.version_window.changes.visualItemRect(item).center()
        QtTest.QTest.mouseClick(self.version_window.changes.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.keyClick(self.version_window.changes.viewport(), QtCore.Qt.Key_Return)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)

    def _change_version_filter(self, index):
        assert self.version_window is not None
        assert index < self.version_window.versionFilterCB.count()
        self.version_window.versionFilterCB.setCurrentIndex(index)
        self.version_window.versionFilterCB.currentIndexChanged.emit(index)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)

    def _set_version_name(self):
        self._change_version_filter(1)
        # make a changes
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        self.version_window.load_all_changes()
        QtWidgets.QApplication.processEvents()
        self._activate_change_at_index(0)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseClick(self.version_window.nameVersionBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_mscolab_merge_waypoints
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module is used to test mscolab-operation related gui.

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import sys
import fs
import mock
import pytest

from mslib.msui import flighttrack as ft
from mslib.mscolab.conf import mscolab_settings
from PyQt5 import QtCore, QtTest, QtWidgets
from mslib._tests.utils import (mscolab_start_server, mscolab_register_and_login, mscolab_create_operation,
                                mscolab_delete_all_operations, mscolab_delete_user)
from mslib.msui import mscolab
import mslib.msui.mss_pyui as mss_pyui
from mslib.mscolab.mscolab import handle_db_reset


PORTS = list(range(23000, 23500))


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_Mscolab_Merge_Waypoints(object):
    def setup(self):
        handle_db_reset()
        self.process, self.url, self.app, _, self.cm, self.fm = mscolab_start_server(PORTS)
        QtTest.QTest.qWait(500)
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = mss_pyui.MSSMainWindow(mscolab_data_dir=mscolab_settings.MSCOLAB_DATA_DIR)
        self.emailid = 'merge@alpha.org'

    def teardown(self):
        with self.app.app_context():
            mscolab_delete_all_operations(self.app, self.url, self.emailid, 'abcdef', 'alpha')
            mscolab_delete_user(self.app, self.url, self.emailid, 'abcdef')
        with fs.open_fs(mscolab_settings.MSCOLAB_DATA_DIR) as mss_dir:
            if mss_dir.exists('local_mscolab_data'):
                mss_dir.removetree('local_mscolab_data')
            assert mss_dir.exists('local_mscolab_data') is False
        if self.window.mscolab.version_window:
            self.window.mscolab.version_window.close()
        if self.window.mscolab.conn:
            self.window.mscolab.conn.disconnect()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        self.process.terminate()

    def _create_user_data(self, emailid='merge@alpha.org'):
        with self.app.app_context():
            self._connect_to_mscolab()
            response = mscolab_register_and_login(self.app, self.url, emailid, 'abcdef', 'alpha')

            assert response.status == '200 OK'
            data, response = mscolab_create_operation(self.app, self.url, response,
                                                      path='f3', description='f3 test example')
            assert response.status == '200 OK'
            self._login(emailid, 'abcdef')
            self._activate_operation_at_index(0)

    def _connect_to_mscolab(self):
        self.connect_window = mscolab.MSColab_ConnectDialog(parent=self.window, mscolab=self.window.mscolab)
        self.window.mscolab.connect_window = self.connect_window
        self.connect_window.urlCb.setEditText(self.url)
        self.connect_window.show()
        QtTest.QTest.mouseClick(self.connect_window.connectBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(500)

    def _login(self, emailid="merge_waypoints_user", password="password"):
        self.connect_window.loginEmailLe.setText(emailid)
        self.connect_window.loginPasswordLe.setText(password)
        QtTest.QTest.mouseClick(self.connect_window.loginBtn, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(500)

    def _activate_operation_at_index(self, index):
        item = self.window.listOperationsMSC.item(index)
        point = self.window.listOperationsMSC.visualItemRect(item).center()
        QtTest.QTest.mouseClick(self.window.listOperationsMSC.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseDClick(self.window.listOperationsMSC.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()

    def _select_waypoints(self, table):
        for row in range(table.model().rowCount()):
            table.selectRow(row)
            QtWidgets.QApplication.processEvents()


@pytest.mark.skip("timeout on github")
class Test_Overwrite_To_Server(Test_Mscolab_Merge_Waypoints):
    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_save_overwrite_to_server(self, mockbox):
        self.emailid = "save_overwrite@alpha.org"
        self._create_user_data(emailid=self.emailid)
        wp_server_before = self.window.mscolab.waypoints_model.waypoint_data(0)
        self.window.workLocallyCheckbox.setChecked(True)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        wp_local = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_local.lat == wp_server_before.lat
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        wp_local_before = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_server_before.lat != wp_local_before.lat

        # ToDo mock this messagebox
        def handle_merge_dialog():
            QtTest.QTest.mouseClick(self.window.mscolab.merge_dialog.overwriteBtn, QtCore.Qt.LeftButton)
            QtWidgets.QApplication.processEvents()
            QtTest.QTest.qWait(100)

        QtCore.QTimer.singleShot(3000, handle_merge_dialog)
        # trigger save to server action from server options combobox
        self.window.serverOptionsCb.setCurrentIndex(2)
        QtWidgets.QApplication.processEvents()
        # get the updated waypoints model from the server
        # ToDo understand why requesting in follow up test self.window.waypoints_model not working
        server_xml = self.window.mscolab.request_wps_from_server()
        server_waypoints_model = ft.WaypointsTableModel(xml_content=server_xml)
        new_local_wp = server_waypoints_model.waypoint_data(0)
        assert wp_local_before.lat == new_local_wp.lat
        self.window.workLocallyCheckbox.setChecked(False)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        new_server_wp = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_local_before.lat == new_server_wp.lat


class Test_Save_Keep_Server_Points(Test_Mscolab_Merge_Waypoints):
    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_save_keep_server_points(self, mockbox):
        self.emailid = "save_keepe@alpha.org"
        self._create_user_data(emailid=self.emailid)
        wp_server_before = self.window.mscolab.waypoints_model.waypoint_data(0)
        self.window.workLocallyCheckbox.setChecked(True)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        wp_local = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_local.lat == wp_server_before.lat
        self.window.mscolab.waypoints_model.invert_direction()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        wp_local_before = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_server_before.lat != wp_local_before.lat

        # ToDo mock this messagebox
        def handle_merge_dialog():
            QtTest.QTest.mouseClick(self.window.mscolab.merge_dialog.keepServerBtn, QtCore.Qt.LeftButton)
            QtWidgets.QApplication.processEvents()
            QtTest.QTest.qWait(100)

        QtCore.QTimer.singleShot(3000, handle_merge_dialog)
        # trigger save to server action from server options combobox
        self.window.serverOptionsCb.setCurrentIndex(2)
        QtWidgets.QApplication.processEvents()
        # get the updated waypoints model from the server
        # ToDo understand why requesting in follow up test self.window.waypoints_model not working
        server_xml = self.window.mscolab.request_wps_from_server()
        server_waypoints_model = ft.WaypointsTableModel(xml_content=server_xml)
        new_local_wp = server_waypoints_model.waypoint_data(0)

        assert wp_local_before.lat != new_local_wp.lat
        assert new_local_wp.lat == wp_server_before.lat
        self.window.workLocallyCheckbox.setChecked(False)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        new_server_wp = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_server_before.lat == new_server_wp.lat


class Test_Fetch_From_Server(Test_Mscolab_Merge_Waypoints):
    @mock.patch("PyQt5.QtWidgets.QMessageBox")
    def test_fetch_from_server(self, mockbox):
        self.emailid = "fetch_from_server@alpha.org"
        self._create_user_data(emailid=self.emailid)
        wp_server_before = self.window.mscolab.waypoints_model.waypoint_data(0)
        self.window.workLocallyCheckbox.setChecked(True)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        wp_local = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_local.lat == wp_server_before.lat
        self.window.mscolab.waypoints_model.invert_direction()
        wp_local_before = self.window.mscolab.waypoints_model.waypoint_data(0)
        assert wp_server_before.lat != wp_local_before.lat

        # ToDo mock this messagebox
        def handle_merge_dialog():
            QtTest.QTest.mouseClick(self.window.mscolab.merge_dialog.keepServerBtn, QtCore.Qt.LeftButton)
            QtWidgets.QApplication.processEvents()
            QtTest.QTest.qWait(100)

        QtCore.QTimer.singleShot(3000, handle_merge_dialog)
        # trigger save to server action from server options combobox
        self.window.serverOptionsCb.setCurrentIndex(1)
        QtWidgets.QApplication.processEvents()
        # get the updated waypoints model from the server
        # ToDo understand why requesting in follow up test of self.window.waypoints_model not working
        server_xml = self.window.mscolab.request_wps_from_server()
        server_waypoints_model = ft.WaypointsTableModel(xml_content=server_xml)
        new_local_wp = server_waypoints_model
        assert len(new_local_wp.waypoints) == 2
        assert new_local_wp.waypoint_data(0).lat == wp_server_before.lat
        self.window.workLocallyCheckbox.setChecked(False)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWait(100)
        assert self.window.mscolab.waypoints_model.waypoint_data(0).lat == wp_server_before.lat



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_configuration_editor_window.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ConfigurationEditorWindow(object):
    def setupUi(self, ConfigurationEditorWindow):
        ConfigurationEditorWindow.setObjectName("ConfigurationEditorWindow")
        ConfigurationEditorWindow.resize(639, 676)
        self.centralwidget = QtWidgets.QWidget(ConfigurationEditorWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setObjectName("label")
        self.horizontalLayout_4.addWidget(self.label)
        self.optCb = QtWidgets.QComboBox(self.centralwidget)
        self.optCb.setObjectName("optCb")
        self.horizontalLayout_4.addWidget(self.optCb)
        self.addOptBtn = QtWidgets.QPushButton(self.centralwidget)
        self.addOptBtn.setObjectName("addOptBtn")
        self.horizontalLayout_4.addWidget(self.addOptBtn)
        self.removeOptBtn = QtWidgets.QPushButton(self.centralwidget)
        self.removeOptBtn.setObjectName("removeOptBtn")
        self.horizontalLayout_4.addWidget(self.removeOptBtn)
        self.restoreDefaultsBtn = QtWidgets.QPushButton(self.centralwidget)
        self.restoreDefaultsBtn.setObjectName("restoreDefaultsBtn")
        self.horizontalLayout_4.addWidget(self.restoreDefaultsBtn)
        self.moveUpTb = QtWidgets.QToolButton(self.centralwidget)
        self.moveUpTb.setObjectName("moveUpTb")
        self.horizontalLayout_4.addWidget(self.moveUpTb)
        self.moveDownTb = QtWidgets.QToolButton(self.centralwidget)
        self.moveDownTb.setObjectName("moveDownTb")
        self.horizontalLayout_4.addWidget(self.moveDownTb)
        self.horizontalLayout_4.setStretch(1, 1)
        self.horizontalLayout_4.setStretch(2, 1)
        self.horizontalLayout_4.setStretch(3, 1)
        self.horizontalLayout_4.setStretch(4, 1)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.jsonWidget = QtWidgets.QWidget(self.centralwidget)
        self.jsonWidget.setObjectName("jsonWidget")
        self.verticalLayout.addWidget(self.jsonWidget)
        self.verticalLayout.setStretch(1, 1)
        self.actionCloseWindow = QtWidgets.QAction(ConfigurationEditorWindow)
        self.actionCloseWindow.setObjectName("actionCloseWindow")
        ConfigurationEditorWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(ConfigurationEditorWindow)
        self.statusbar.setObjectName("statusbar")
        ConfigurationEditorWindow.setStatusBar(self.statusbar)
        self.toolBar = QtWidgets.QToolBar(ConfigurationEditorWindow)
        self.toolBar.setObjectName("toolBar")
        ConfigurationEditorWindow.addToolBar(QtCore.Qt.TopToolBarArea, self.toolBar)
        self.centralwidget.addAction(self.actionCloseWindow)

        self.retranslateUi(ConfigurationEditorWindow)
        self.actionCloseWindow.triggered.connect(ConfigurationEditorWindow.close)
        QtCore.QMetaObject.connectSlotsByName(ConfigurationEditorWindow)

    def retranslateUi(self, ConfigurationEditorWindow):
        _translate = QtCore.QCoreApplication.translate
        ConfigurationEditorWindow.setWindowTitle(_translate("ConfigurationEditorWindow", "MSS Configuration Editor"))
        self.label.setText(_translate("ConfigurationEditorWindow", "Filter :"))
        self.optCb.setToolTip(_translate("ConfigurationEditorWindow", "Select option to filter the view"))
        self.addOptBtn.setToolTip(_translate("ConfigurationEditorWindow", "Select an option to add new value"))
        self.addOptBtn.setText(_translate("ConfigurationEditorWindow", "+ Add"))
        self.removeOptBtn.setToolTip(_translate("ConfigurationEditorWindow", "Select one/more options to remove"))
        self.removeOptBtn.setText(_translate("ConfigurationEditorWindow", "- Remove"))
        self.restoreDefaultsBtn.setToolTip(_translate("ConfigurationEditorWindow", "Select one/more options to restore default value"))
        self.restoreDefaultsBtn.setText(_translate("ConfigurationEditorWindow", "Restore Defaults"))
        self.moveUpTb.setText(_translate("ConfigurationEditorWindow", "..."))
        self.moveDownTb.setText(_translate("ConfigurationEditorWindow", "..."))
        self.actionCloseWindow.setText(_translate("ConfigurationEditorWindow", "CloseWindow"))
        self.actionCloseWindow.setShortcut(_translate("ConfigurationEditorWindow", "Ctrl+W"))
        self.toolBar.setWindowTitle(_translate("ConfigurationEditorWindow", "toolBar"))



# -*- coding: utf-8 -*-
"""

    mslib.plugins.io.flitestar
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    plugin for flitestar format flight track export

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import division

import numpy as np
import os
from fs import open_fs

import mslib.msui.flighttrack as ft
from mslib.utils import thermolib
from mslib.utils.units import units


def load_from_flitestar(filename):
    waypoints = []
    _dirname, _name = os.path.split(filename)
    _fs = open_fs(_dirname)
    with _fs.open(_name, 'r') as f:
        firstline = f.readline()
        if not firstline.startswith("# FliteStar/FliteMap generated flight plan."):
            raise SyntaxError("The file does not seem to be a FliteStar file!")
        for line in f:

            if line.startswith('FWP'):
                line = line.split()
                if len(line) < 10:
                    raise SyntaxError(f"Line {line} has less than 9 fields.")
                alt = round(float(line[-1]) / 100., 2)
                if line[4] == 'N':
                    NS = 1.
                elif line[4] == 'S':
                    NS = -1.
                else:
                    NS = np.nan
                lat = round((float(line[5]) + (float(line[6]) / 60.)) * NS, 3)
                if line[7] == 'E':
                    EW = 1.
                elif line[7] == 'W':
                    EW = -1.
                else:
                    EW = np.nan
                lon = round((float(line[8]) + (float(line[9]) / 60.)) * EW, 3)

                wp = ft.Waypoint()
                wp.location = line[3]
                wp.lat = float(lat)
                wp.lon = float(lon)
                wp.flightlevel = float(alt)
                wp.pressure = thermolib.flightlevel2pressure(float(wp.flightlevel) * units.hft).magnitude
                waypoints.append(wp)

    name = os.path.basename(filename).strip('.txt')
    return name, waypoints



# -*- coding: utf-8 -*-
"""

    mslib.utils._tests.test_time
    ~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests mslib.utils.time

    This file is part of mss.

    :copyright: Copyright 2016-2017 Reimar Bauer
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
import datetime
import mslib.utils.time as time

LOGGER = logging.getLogger(__name__)


class TestParseTime(object):
    def test_parse_iso_datetime(self):
        assert time.parse_iso_datetime("2009-05-28T16:15:00") == datetime.datetime(2009, 5, 28, 16, 15)

    def test_parse_iso_duration(self):
        assert time.parse_iso_duration('P01W') == datetime.timedelta(days=7)


class TestTimes(object):
    """
    tests about times
    """

    def test_datetime_to_jsec(self):
        assert time.datetime_to_jsec(datetime.datetime(2000, 2, 1, 0, 0, 0, 0)) == 2678400.0
        assert time.datetime_to_jsec(datetime.datetime(2000, 1, 1, 0, 0, 0, 0)) == 0
        assert time.datetime_to_jsec(datetime.datetime(1995, 1, 1, 0, 0, 0, 0)) == -157766400.0

    def test_jsec_to_datetime(self):
        assert time.jsec_to_datetime(0) == datetime.datetime(2000, 1, 1, 0, 0, 0, 0)
        assert time.jsec_to_datetime(3600) == datetime.datetime(2000, 1, 1, 1, 0, 0, 0)
        assert time.jsec_to_datetime(-157766400.0) == datetime.datetime(1995, 1, 1, 0, 0, 0, 0)

    def test_compute_hour_of_day(self):
        assert time.compute_hour_of_day(0) == 0
        assert time.compute_hour_of_day(86400) == 0
        assert time.compute_hour_of_day(3600) == 1
        assert time.compute_hour_of_day(82800) == 23



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/ui_remotesensing_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_RemoteSensingDockWidget(object):
    def setupUi(self, RemoteSensingDockWidget):
        RemoteSensingDockWidget.setObjectName("RemoteSensingDockWidget")
        RemoteSensingDockWidget.resize(465, 146)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(RemoteSensingDockWidget.sizePolicy().hasHeightForWidth())
        RemoteSensingDockWidget.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(RemoteSensingDockWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.lbObsAngle = QtWidgets.QLabel(RemoteSensingDockWidget)
        self.lbObsAngle.setObjectName("lbObsAngle")
        self.horizontalLayout.addWidget(self.lbObsAngle)
        self.dsbObsAngleAzimuth = QtWidgets.QDoubleSpinBox(RemoteSensingDockWidget)
        self.dsbObsAngleAzimuth.setDecimals(0)
        self.dsbObsAngleAzimuth.setMinimum(-180.0)
        self.dsbObsAngleAzimuth.setMaximum(180.0)
        self.dsbObsAngleAzimuth.setSingleStep(15.0)
        self.dsbObsAngleAzimuth.setObjectName("dsbObsAngleAzimuth")
        self.horizontalLayout.addWidget(self.dsbObsAngleAzimuth)
        self.label = QtWidgets.QLabel(RemoteSensingDockWidget)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.dsbObsAngleElevation = QtWidgets.QDoubleSpinBox(RemoteSensingDockWidget)
        self.dsbObsAngleElevation.setMinimum(-90.0)
        self.dsbObsAngleElevation.setMaximum(90.0)
        self.dsbObsAngleElevation.setObjectName("dsbObsAngleElevation")
        self.horizontalLayout.addWidget(self.dsbObsAngleElevation)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.cbDrawTangents = QtWidgets.QCheckBox(RemoteSensingDockWidget)
        self.cbDrawTangents.setMinimumSize(QtCore.QSize(145, 0))
        self.cbDrawTangents.setObjectName("cbDrawTangents")
        self.horizontalLayout_5.addWidget(self.cbDrawTangents)
        self.btTangentsColour = QtWidgets.QPushButton(RemoteSensingDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btTangentsColour.sizePolicy().hasHeightForWidth())
        self.btTangentsColour.setSizePolicy(sizePolicy)
        self.btTangentsColour.setMinimumSize(QtCore.QSize(135, 0))
        self.btTangentsColour.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.btTangentsColour.setObjectName("btTangentsColour")
        self.horizontalLayout_5.addWidget(self.btTangentsColour)
        self.dsbTangentHeight = QtWidgets.QDoubleSpinBox(RemoteSensingDockWidget)
        self.dsbTangentHeight.setMinimumSize(QtCore.QSize(0, 0))
        self.dsbTangentHeight.setPrefix("")
        self.dsbTangentHeight.setObjectName("dsbTangentHeight")
        self.horizontalLayout_5.addWidget(self.dsbTangentHeight)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.cbShowSolarAngle = QtWidgets.QCheckBox(RemoteSensingDockWidget)
        self.cbShowSolarAngle.setMinimumSize(QtCore.QSize(145, 0))
        self.cbShowSolarAngle.setObjectName("cbShowSolarAngle")
        self.horizontalLayout_6.addWidget(self.cbShowSolarAngle)
        self.cbSolarAngleType = QtWidgets.QComboBox(RemoteSensingDockWidget)
        self.cbSolarAngleType.setObjectName("cbSolarAngleType")
        self.cbSolarAngleType.addItem("")
        self.cbSolarAngleType.addItem("")
        self.cbSolarAngleType.addItem("")
        self.horizontalLayout_6.addWidget(self.cbSolarAngleType)
        self.cbSolarBody = QtWidgets.QComboBox(RemoteSensingDockWidget)
        self.cbSolarBody.setMinimumSize(QtCore.QSize(170, 0))
        self.cbSolarBody.setObjectName("cbSolarBody")
        self.cbSolarBody.addItem("")
        self.cbSolarBody.addItem("")
        self.cbSolarBody.addItem("")
        self.cbSolarBody.addItem("")
        self.horizontalLayout_6.addWidget(self.cbSolarBody)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.lbSolarCmap = QtWidgets.QLabel(RemoteSensingDockWidget)
        self.lbSolarCmap.setObjectName("lbSolarCmap")
        self.horizontalLayout_2.addWidget(self.lbSolarCmap)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem3)

        self.retranslateUi(RemoteSensingDockWidget)
        QtCore.QMetaObject.connectSlotsByName(RemoteSensingDockWidget)

    def retranslateUi(self, RemoteSensingDockWidget):
        _translate = QtCore.QCoreApplication.translate
        RemoteSensingDockWidget.setWindowTitle(_translate("RemoteSensingDockWidget", "Remote Sensing"))
        self.lbObsAngle.setToolTip(_translate("RemoteSensingDockWidget", "View direction of the remote sensing instrument.\n"
"0 degree is towards flight direction."))
        self.lbObsAngle.setText(_translate("RemoteSensingDockWidget", "Viewing direction: azimuth"))
        self.label.setText(_translate("RemoteSensingDockWidget", "elevation"))
        self.cbDrawTangents.setToolTip(_translate("RemoteSensingDockWidget", "Tangent points in viewing direction at the specified altitude.\n"
"Aircraft altitude is taken from the flight path."))
        self.cbDrawTangents.setText(_translate("RemoteSensingDockWidget", "draw tangent points"))
        self.btTangentsColour.setText(_translate("RemoteSensingDockWidget", "colour"))
        self.dsbTangentHeight.setSuffix(_translate("RemoteSensingDockWidget", " km"))
        self.cbShowSolarAngle.setToolTip(_translate("RemoteSensingDockWidget", "dark green if below horizon; otherwise reds: 0,5,10,15, purples: 15,25,35,45,60, greens: 60,90,135,180"))
        self.cbShowSolarAngle.setText(_translate("RemoteSensingDockWidget", "show angle (degree)"))
        self.cbSolarAngleType.setItemText(0, _translate("RemoteSensingDockWidget", "sun"))
        self.cbSolarAngleType.setItemText(1, _translate("RemoteSensingDockWidget", "moon"))
        self.cbSolarAngleType.setItemText(2, _translate("RemoteSensingDockWidget", "venus"))
        self.cbSolarBody.setItemText(0, _translate("RemoteSensingDockWidget", "total (horizon)"))
        self.cbSolarBody.setItemText(1, _translate("RemoteSensingDockWidget", "total"))
        self.cbSolarBody.setItemText(2, _translate("RemoteSensingDockWidget", "azimuth"))
        self.cbSolarBody.setItemText(3, _translate("RemoteSensingDockWidget", "elevation"))
        self.lbSolarCmap.setText(_translate("RemoteSensingDockWidget", "fill me"))




# -*- coding: utf-8 -*-
"""

    mslib.mswms._tests.utils
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides utils for pytest to test mslib modules

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import requests
import time
import fs
import socket
import multiprocessing

from flask_testing import LiveServerTestCase

from PyQt5 import QtTest
from werkzeug.urls import url_join
from mslib.mscolab.server import register_user
from flask import json
from mslib._tests.constants import MSS_CONFIG_PATH
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.server import APP, initialize_managers, start_server
from mslib.mscolab.mscolab import handle_db_init


def callback_ok_image(status, response_headers):
    assert status == "200 OK"
    assert response_headers[0] == ('Content-type', 'image/png')


def callback_ok_xml(status, response_headers):
    assert status == "200 OK"
    assert response_headers[0] == ('Content-type', 'text/xml')


def callback_ok_html(status, response_headers):
    assert status == "200 OK"
    assert response_headers[0] == ('Content-Type', 'text/html; charset=utf-8')


def callback_404_plain(status, response_headers):
    assert status == "404 NOT FOUND"
    assert response_headers[0] == ('Content-type', 'text/plain')


def callback_307_html(status, response_headers):
    assert status == "307 TEMPORARY REDIRECT"
    assert response_headers[0] == ('Content-Type', 'text/html; charset=utf-8')


def mscolab_register_user(app, msc_url, email, password, username):
    # Duplicate of imported register_user
    data = {
        'email': email,
        'password': password,
        'username': username
    }
    url = url_join(msc_url, 'register')
    response = app.test_client().post(url, data=data)
    return response


def mscolab_register_and_login(app, msc_url, email, password, username):
    register_user(email, password, username)
    data = {
        'email': email,
        'password': password
    }
    url = url_join(msc_url, 'token')
    response = app.test_client().post(url, data=data)
    return response


def mscolab_login(app, msc_url, email='a', password='a'):
    data = {
        'email': email,
        'password': password
    }
    url = url_join(msc_url, 'token')
    response = app.test_client().post(url, data=data)
    return response


def mscolab_delete_user(app, msc_url, email, password):
    with app.app_context():
        response = mscolab_login(app, msc_url, email, password)
        if response.status == '200 OK':
            data = json.loads(response.get_data(as_text=True))
            url = url_join(msc_url, 'delete_user')
            response = app.test_client().post(url, data=data)
            if response.status == '200 OK':
                data = json.loads(response.get_data(as_text=True))
                return data["success"]
        return False


def mscolab_create_content(app, msc_url, data, path_name='example', content=None):
    if content is None:
        content = f"""\
        <?xml version="1.0" encoding="utf-8"?>
          <FlightTrack>
            <Name>{path_name}</Name>
            <ListOfWaypoints>
              <Waypoint flightlevel="0.0" lat="55.15" location="A1" lon="-23.74">
                <Comments>Takeoff</Comments>
              </Waypoint>
              <Waypoint flightlevel="350" lat="42.99" location="A2" lon="-12.1">
                <Comments></Comments>
              </Waypoint>
              </ListOfWaypoints>
          </FlightTrack>"""
    data["path"] = path_name
    data['description'] = path_name
    data['content'] = content
    url = url_join(msc_url, 'create_operation')
    response = app.test_client().post(url, data=data)
    return response


def mscolab_delete_all_operations(app, msc_url, email, password, username):
    response = mscolab_register_and_login(app, msc_url, email, password, username)
    data = json.loads(response.get_data(as_text=True))
    url = url_join(msc_url, 'operations')
    response = app.test_client().get(url, data=data)
    response = json.loads(response.get_data(as_text=True))
    for p in response['operations']:
        data['op_id'] = p['op_id']
        url = url_join(msc_url, 'delete_operation')
        response = app.test_client().post(url, data=data)


def mscolab_create_operation(app, msc_url, response, path='f', description='description'):
    data = json.loads(response.get_data(as_text=True))
    data["path"] = path
    data['description'] = description
    url = url_join(msc_url, 'create_operation')
    response = app.test_client().post(url, data=data)
    return data, response


def mscolab_get_operation_id(app, msc_url, email, password, username, path):
    response = mscolab_register_and_login(app, msc_url, email, password, username)
    data = json.loads(response.get_data(as_text=True))
    url = url_join(msc_url, 'operations')
    response = app.test_client().get(url, data=data)
    response = json.loads(response.get_data(as_text=True))
    for p in response['operations']:
        if p['path'] == path:
            return p['op_id']


def mscolab_check_free_port(all_ports, port):
    _s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        _s.bind(("127.0.0.1", port))
    except (socket.error, IOError):
        port = all_ports.pop()
        port = mscolab_check_free_port(all_ports, port)
    else:
        _s.close()
    return port


def mscolab_ping_server(port):
    url = f"http://127.0.0.1:{port}/status"
    try:
        r = requests.get(url)
        if r.text == "Mscolab server":
            return True
    except requests.exceptions.ConnectionError:
        return False
    return False


def mscolab_start_server(all_ports, mscolab_settings=mscolab_settings, timeout=5):
    handle_db_init()
    port = mscolab_check_free_port(all_ports, all_ports.pop())

    url = f"http://localhost:{port}"

    _app = APP
    _app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
    _app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
    _app.config['UPLOAD_FOLDER'] = mscolab_settings.UPLOAD_FOLDER
    _app.config['URL'] = url

    _app, sockio, cm, fm = initialize_managers(_app)

    # ToDo refactoring for spawn needed, fork is not implemented on windows, spawn is default on MAC and Windows
    if multiprocessing.get_start_method(allow_none=True) != 'fork':
        multiprocessing.set_start_method("fork")
    process = multiprocessing.Process(
        target=start_server,
        args=(_app, sockio, cm, fm,),
        kwargs={'port': port})
    process.start()
    start_time = time.time()
    while True:
        elapsed_time = (time.time() - start_time)
        if elapsed_time > timeout:
            raise RuntimeError(
                "Failed to start the server after %d seconds. " % timeout
            )

        if mscolab_ping_server(port):
            break

    return process, url, _app, sockio, cm, fm


def create_mss_settings_file(content):
    with fs.open_fs(MSS_CONFIG_PATH) as file_dir:
        file_dir.writetext("mss_settings.json", content)


def wait_until_signal(signal, timeout=5):
    """
    Blocks the calling thread until the signal emits or the timeout expires.
    """
    init_time = time.time()
    finished = False

    def done(*args):
        nonlocal finished
        finished = True

    signal.connect(done)
    while not finished and time.time() - init_time < timeout:
        QtTest.QTest.qWait(100)

    try:
        signal.disconnect(done)
    except TypeError:
        pass
    finally:
        return finished


class ExceptionMock:
    """
    Replace function calls with raised exceptions
    e.g.
    with mock.patch("requests.get", new=ExceptionMock(requests.exceptions.ConnectionError).raise_exc):
        self._login()
    """
    def __init__(self, exc):
        self.exc = exc

    def raise_exc(self, *args, **kwargs):
        raise self.exc


class LiveSocketTestCase(LiveServerTestCase):

    def _spawn_live_server(self):
        self._process = None
        port_value = self._port_value
        app, sockio, cm, fm = initialize_managers(self.app)
        self._process = multiprocessing.Process(
            target=start_server,
            args=(app, sockio, cm, fm,),
            kwargs={'port': port_value.value})

        self._process.start()

        # We must wait for the server to start listening, but give up
        # after a specified maximum timeout
        timeout = self.app.config.get('LIVESERVER_TIMEOUT', 5)
        start_time = time.time()

        while True:
            elapsed_time = (time.time() - start_time)
            if elapsed_time > timeout:
                raise RuntimeError(
                    "Failed to start the server after %d seconds. " % timeout
                )

            if self._can_ping_server():
                break



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_wms_capabilities.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_WMSCapabilitiesBrowser(object):
    def setupUi(self, WMSCapabilitiesBrowser):
        WMSCapabilitiesBrowser.setObjectName("WMSCapabilitiesBrowser")
        WMSCapabilitiesBrowser.resize(754, 393)
        self.verticalLayout = QtWidgets.QVBoxLayout(WMSCapabilitiesBrowser)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(WMSCapabilitiesBrowser)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.lblURL = QtWidgets.QLabel(WMSCapabilitiesBrowser)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblURL.sizePolicy().hasHeightForWidth())
        self.lblURL.setSizePolicy(sizePolicy)
        self.lblURL.setObjectName("lblURL")
        self.horizontalLayout_2.addWidget(self.lblURL)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.txtCapabilities = QtWidgets.QTextBrowser(WMSCapabilitiesBrowser)
        self.txtCapabilities.setOpenExternalLinks(False)
        self.txtCapabilities.setObjectName("txtCapabilities")
        self.verticalLayout.addWidget(self.txtCapabilities)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.cbFullView = QtWidgets.QCheckBox(WMSCapabilitiesBrowser)
        self.cbFullView.setObjectName("cbFullView")
        self.horizontalLayout.addWidget(self.cbFullView)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btClose = QtWidgets.QPushButton(WMSCapabilitiesBrowser)
        self.btClose.setObjectName("btClose")
        self.horizontalLayout.addWidget(self.btClose)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(WMSCapabilitiesBrowser)
        self.btClose.clicked.connect(WMSCapabilitiesBrowser.close)
        QtCore.QMetaObject.connectSlotsByName(WMSCapabilitiesBrowser)

    def retranslateUi(self, WMSCapabilitiesBrowser):
        _translate = QtCore.QCoreApplication.translate
        WMSCapabilitiesBrowser.setWindowTitle(_translate("WMSCapabilitiesBrowser", "Browse WMS Capabilities - Mission Support System"))
        self.label.setText(_translate("WMSCapabilitiesBrowser", "Capabilities document of:"))
        self.lblURL.setText(_translate("WMSCapabilitiesBrowser", "<URL>"))
        self.cbFullView.setText(_translate("WMSCapabilitiesBrowser", "show full XML document"))
        self.btClose.setText(_translate("WMSCapabilitiesBrowser", "close"))




# -*- coding: utf-8 -*-
"""

    mslib.msui.remotesensing_dockwidget
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Control widget to configure remote sensing overlays.

    This file is part of mss.

    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2017-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import collections

from matplotlib.collections import LineCollection
from matplotlib.colors import BoundaryNorm, ListedColormap
import numpy as np
from skyfield.api import Loader, Topos, utc
import skyfield_data

from PyQt5 import QtGui, QtWidgets
from mslib.msui.mss_qt import ui_remotesensing_dockwidget as ui
from mslib.utils.time import jsec_to_datetime, datetime_to_jsec
from mslib.utils.coordinate import get_distance, rotate_point, fix_angle


EARTH_RADIUS = 6371.


class RemoteSensingControlWidget(QtWidgets.QWidget, ui.Ui_RemoteSensingDockWidget):
    """This class implements the remote sensing functionality as dockable widget.
    """

    def __init__(self, parent=None, view=None):
        """
        Arguments:
        parent -- Qt widget that is parent to this widget.
        view -- reference to mpl canvas class
        """
        super(RemoteSensingControlWidget, self).__init__(parent)
        self.setupUi(self)

        self.view = view
        self.load_bsp = Loader(skyfield_data.get_skyfield_data_path(), verbose=False)

        self.planets = self.load_bsp('de421.bsp')
        self.timescale = self.load_bsp.timescale(builtin=True)

        button = self.btTangentsColour
        palette = QtGui.QPalette(button.palette())
        colour = QtGui.QColor()
        colour.setRgbF(1, 0, 0, 1)
        palette.setColor(QtGui.QPalette.Button, colour)
        button.setPalette(palette)

        self.dsbTangentHeight.setValue(10.)
        self.dsbObsAngleAzimuth.setValue(90.)
        self.dsbObsAngleElevation.setValue(-1.0)

        # update plot on every value change
        self.cbDrawTangents.stateChanged.connect(self.update_settings)
        self.cbShowSolarAngle.stateChanged.connect(self.update_settings)
        self.btTangentsColour.clicked.connect(self.set_tangentpoint_colour)
        self.dsbTangentHeight.valueChanged.connect(self.update_settings)
        self.dsbObsAngleAzimuth.valueChanged.connect(self.update_settings)
        self.dsbObsAngleElevation.valueChanged.connect(self.update_settings)
        self.cbSolarBody.currentIndexChanged.connect(self.update_settings)
        self.cbSolarAngleType.currentIndexChanged.connect(self.update_settings)
        self.lbSolarCmap.setText(
            "Solar angle colours, dark to light: reds (0-15), violets (15-45), greens (45-180)")
        self.solar_cmap = ListedColormap([
            (1.00, 0.00, 0.00, 1.0),
            (1.00, 0.45, 0.00, 1.0),
            (1.00, 0.75, 0.00, 1.0),
            (0.47, 0.10, 1.00, 1.0),
            (0.72, 0.38, 1.00, 1.0),
            (1.00, 0.55, 1.00, 1.0),
            (0.00, 0.70, 0.00, 1.0),
            (0.33, 0.85, 0.33, 1.0),
            (0.65, 1.00, 0.65, 1.0)])
        self.solar_norm = BoundaryNorm(
            [0, 5, 10, 15, 25, 35, 45, 90, 135, 180], self.solar_cmap.N)

        self.update_settings()

    @staticmethod
    def compute_view_angles(lon0, lat0, h0, lon1, lat1, h1, obs_azi, obs_ele):
        mlat = ((lat0 + lat1) / 2.)
        lon0 *= np.cos(np.deg2rad(mlat))
        lon1 *= np.cos(np.deg2rad(mlat))
        dlon = lon1 - lon0
        dlat = lat1 - lat0
        obs_azi_p = fix_angle(obs_azi + np.rad2deg(np.arctan2(dlon, dlat)))
        return obs_azi_p, obs_ele

    def compute_body_angle(self, body, jsec, lon, lat):
        t = self.timescale.utc(jsec_to_datetime(jsec).replace(tzinfo=utc))
        loc = self.planets["earth"] + Topos(lat, lon)
        astrometric = loc.at(t).observe(self.planets[body])
        alt, az, d = astrometric.apparent().altaz()
        return az.degrees, alt.degrees

    def update_settings(self):
        """
        Updates settings in TopView and triggers a redraw.
        """
        settings = {
            "reference": self,
            "draw_tangents": self.cbDrawTangents.isChecked(),
        }
        if self.cbShowSolarAngle.isChecked():
            settings["show_solar_angle"] = self.cbSolarAngleType.currentText(), self.cbSolarBody.currentText()
        else:
            settings["show_solar_angle"] = None

        self.view.set_remote_sensing_appearance(settings)

    def set_tangentpoint_colour(self):
        """Slot for the colour buttons: Opens a QColorDialog and sets the
           new button face colour.
        """
        button = self.btTangentsColour
        palette = QtGui.QPalette(button.palette())
        colour = palette.color(QtGui.QPalette.Button)
        colour = QtWidgets.QColorDialog.getColor(colour)
        if colour.isValid():
            palette.setColor(QtGui.QPalette.Button, colour)
            button.setPalette(palette)
        self.update_settings()

    def compute_tangent_lines(self, bmap, wp_vertices, wp_heights):
        """
        Computes Tangent points of limb sounders aboard the aircraft

        Args:
            bmap: Projection of TopView
            wp_vertices: waypoints of the flight path
            wp_heights: altitude of the waypoints of flight path

        Returns: LineCollection of dotted lines at tangent point locations
        """
        x, y = list(zip(*wp_vertices))
        wp_lons, wp_lats = bmap(x, y, inverse=True)
        fine_lines = [bmap.gcpoints2(
                      wp_lons[i], wp_lats[i], wp_lons[i + 1], wp_lats[i + 1], del_s=10., map_coords=False)
                      for i in range(len(wp_lons) - 1)]
        line_heights = [np.linspace(wp_heights[i], wp_heights[i + 1], num=len(fine_lines[i][0]))
                        for i in range(len(fine_lines))]
        # fine_lines = list of tuples with x-list and y-list for each segment
        tplines = [self.tangent_point_coordinates(
            fine_lines[i][0], fine_lines[i][1], line_heights[i],
            cut_height=self.dsbTangentHeight.value()) for i in range(len(fine_lines))]
        dirlines = self.direction_coordinates(wp_lons, wp_lats)
        lines = tplines + dirlines
        for i, line in enumerate(lines):
            for j, (lon, lat) in enumerate(line):
                line[j] = bmap(lon, lat)
            lines[i] = line
        return LineCollection(
            lines,
            colors=QtGui.QPalette(self.btTangentsColour.palette()).color(QtGui.QPalette.Button).getRgbF(),
            zorder=2, animated=True, linewidth=3, linestyles=[':'] * len(tplines) + ['-'] * len(dirlines))

    def compute_solar_lines(self, bmap, wp_vertices, wp_heights, wp_times, solartype):
        """
        Computes coloured overlay over the flight path that indicates
        the danger of looking into the sun with a limb sounder aboard
        the aircraft.

        Args:
            bmap: Projection of TopView
            wp_vertices: waypoints of the flight path
            wp_heights: altitude of the waypoints of flight path

        Returns: LineCollection of coloured lines according to the
                 angular distance between viewing direction and solar
                 angle
        """
        # calculate distances and times
        body, difftype = solartype

        times = [datetime_to_jsec(_wp_time) for _wp_time in wp_times]
        x, y = list(zip(*wp_vertices))
        wp_lons, wp_lats = bmap(x, y, inverse=True)

        fine_lines = [bmap.gcpoints2(wp_lons[i], wp_lats[i], wp_lons[i + 1], wp_lats[i + 1], map_coords=False) for i in
                      range(len(wp_lons) - 1)]
        line_heights = [np.linspace(wp_heights[i], wp_heights[i + 1], num=len(fine_lines[i][0])) for i in
                        range(len(fine_lines))]
        line_times = [np.linspace(times[i], times[i + 1], num=len(fine_lines[i][0])) for i in
                      range(len(fine_lines))]
        # fine_lines = list of tuples with x-list and y-list for each segment
        # lines = list of tuples with lon-list and lat-list for each segment
        heights = []
        times = []
        for i in range(len(fine_lines) - 1):
            heights.extend(line_heights[i][:-1])
            times.extend(line_times[i][:-1])
        heights.extend(line_heights[-1])
        times.extend(line_times[-1])
        solar_x = []
        solar_y = []
        for i in range(len(fine_lines) - 1):
            solar_x.extend(fine_lines[i][0][:-1])
            solar_y.extend(fine_lines[i][1][:-1])
        solar_x.extend(fine_lines[-1][0])
        solar_y.extend(fine_lines[-1][1])
        points = []
        old_wp = None
        total_distance = 0
        for i, (lon, lat) in enumerate(zip(solar_x, solar_y)):
            points.append([[lon, lat]])  # append double-list for later concatenation
            if old_wp is not None:
                wp_dist = get_distance(old_wp[0], old_wp[1], lat, lon) * 1000.
                total_distance += wp_dist
            old_wp = (lat, lon)
        vals = []
        for i in range(len(points) - 1):
            p0, p1 = points[i][0], points[i + 1][0]

            sol_azi, sol_ele = self.compute_body_angle(body, times[i], p0[0], p0[1])
            obs_azi, obs_ele = self.compute_view_angles(
                p0[0], p0[1], heights[i], p1[0], p1[1], heights[i + 1],
                self.dsbObsAngleAzimuth.value(), self.dsbObsAngleElevation.value())
            if sol_azi < 0:
                sol_azi += 360
            if obs_azi < 0:
                obs_azi += 360
            rating = self.calc_view_rating(obs_azi, obs_ele, sol_azi, sol_ele, heights[i], difftype)
            vals.append(rating)

        # convert lon, lat to map points
        for i in range(len(points)):
            points[i][0][0], points[i][0][1] = bmap(points[i][0][0], points[i][0][1])
        points = np.concatenate([points[:-1], points[1:]], axis=1)
        # plot
        solar_lines = LineCollection(points, cmap=self.solar_cmap, norm=self.solar_norm,
                                     zorder=2, linewidths=3, animated=True)
        solar_lines.set_array(np.array(vals))
        return solar_lines

    def tangent_point_coordinates(self, lon_lin, lat_lin, flight_alt=14, cut_height=12):
        """
        Computes coordinates of tangent points given coordinates of flight path.

        Args:
            lon_lin: longitudes of flight path
            lat_lin: latitudes of flight path
            flight_alt: altitude of aircraft (scalar or numpy array)
            cut_height: altitude of tangent points

        Returns: List of tuples of longitude/latitude coordinates

        """
        lins = list(zip(lon_lin[0:-1], lon_lin[1:], lat_lin[0:-1], lat_lin[1:]))
        lins = [(x0 * np.cos(np.deg2rad(np.mean([y0, y1]))), x1 * np.cos(np.deg2rad(np.mean([y0, y1]))), y0, y1)
                for x0, x1, y0, y1 in lins]

        direction = [(x1 - x0, y1 - y0) for x0, x1, y0, y1 in lins]
        direction = [(_x / np.hypot(_x, _y), _y / np.hypot(_x, _y))
                     for _x, _y in direction]
        los = [rotate_point(point, -self.dsbObsAngleAzimuth.value()) for point in direction]
        los.append(los[-1])

        if isinstance(flight_alt, (collections.abc.Sequence, np.ndarray)):
            dist = [(np.sqrt(max((EARTH_RADIUS + a) ** 2 - (EARTH_RADIUS + cut_height) ** 2, 0)) / 110.)
                    for a in flight_alt[:-1]]
            dist.append(dist[-1])
        else:
            dist = (np.sqrt((EARTH_RADIUS + flight_alt) ** 2 - (EARTH_RADIUS + cut_height) ** 2) / 110.)

        tp_dir = (np.array(los).T * dist).T

        tps = [(x0 + tp_x, y0 + tp_y, y0) for
               ((x0, x1, y0, y1), (tp_x, tp_y)) in zip(lins, tp_dir)]
        tps = [(x0 / np.cos(np.deg2rad(yp)), y0) for (x0, y0, yp) in tps]
        return tps

    def direction_coordinates(self, lon_lin, lat_lin):
        """
        Computes coordinates of tangent points given coordinates of flight path.

        Args:
            lon_lin: longitudes of flight path
            lat_lin: latitudes of flight path
            flight_alt: altitude of aircraft (scalar or numpy array)
            cut_height: altitude of tangent points

        Returns: List of tuples of longitude/latitude coordinates

        """
        lins = list(zip(lon_lin[0:-1], lon_lin[1:], lat_lin[0:-1], lat_lin[1:]))
        lins = [(x0 * np.cos(np.deg2rad(np.mean([y0, y1]))), x1 * np.cos(np.deg2rad(np.mean([y0, y1]))), y0, y1)
                for x0, x1, y0, y1 in lins]
        lens = [np.hypot(x1 - x0, y1 - y0) * 110. for x0, x1, y0, y1 in lins]
        lins = [_x for _x, _l in zip(lins, lens) if _l > 10]

        direction = [(0.5 * (x0 + x1), 0.5 * (y0 + y1), x1 - x0, y1 - y0) for x0, x1, y0, y1 in lins]
        direction = [(_u, _v, _x / np.hypot(_x, _y), _y / np.hypot(_x, _y))
                     for _u, _v, _x, _y in direction]
        los = [rotate_point(point[2:], -self.dsbObsAngleAzimuth.value()) for point in direction]

        dist = 1.

        tp_dir = (np.array(los).T * dist).T

        tps = [(x0, y0, x0 + tp_x, y0 + tp_y) for
               ((x0, y0, _, _), (tp_x, tp_y)) in zip(direction, tp_dir)]
        tps = [[(x0 / np.cos(np.deg2rad(y0)), y0), (x1 / np.cos(np.deg2rad(y0)), y1)] for (x0, y0, x1, y1) in tps]
        return tps

    @staticmethod
    def calc_view_rating(obs_azi, obs_ele, sol_azi, sol_ele, height, difftype):
        """
        Calculates the angular distance between given directions under the
        condition that the sun is above the horizon.

        Args:
            obs_azi: observator azimuth angle
            obs_ele: observator elevation angle
            sol_azi: solar azimuth angle
            sol_ele: solar elevation angle
            height: altitude of observer

        Returns: angular distance or 180 degrees if sun is below horizon
        """
        delta_azi = obs_azi - sol_azi
        delta_ele = obs_ele - sol_ele
        if "horizon" in difftype:
            thresh = -np.rad2deg(np.arccos(EARTH_RADIUS / (height + EARTH_RADIUS))) - 3
            if sol_ele < thresh:
                delta_ele = 180

        if "azimuth" == difftype:
            return np.abs(obs_azi - sol_azi)
        elif "elevation" == difftype:
            return np.abs(obs_ele - sol_ele)
        else:
            return np.hypot(delta_azi, delta_ele)



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_files_api
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    api integration tests for file based handlers

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from flask_testing import TestCase
import os
import fs
import pytest

from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import Operation, db
from mslib.mscolab.server import APP
from mslib.mscolab.file_manager import FileManager
from mslib.mscolab.seed import add_user, get_user
from mslib.mscolab.mscolab import handle_db_reset


@pytest.mark.skipif(os.name == "nt",
                    reason="multiprocessing needs currently start_method fork")
class Test_Files(TestCase):
    render_templates = False

    def create_app(self):
        app = APP
        app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
        app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
        app.config['UPLOAD_FOLDER'] = mscolab_settings.UPLOAD_FOLDER
        app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
        app.config["TESTING"] = True
        app.config['LIVESERVER_TIMEOUT'] = 10
        app.config['LIVESERVER_PORT'] = 0
        return app

    def setUp(self):
        handle_db_reset()
        db.init_app(self.app)

        self.fm = FileManager(self.app.config["MSCOLAB_DATA_DIR"])
        self.userdata = 'UV10@uv10', 'UV10', 'uv10'

        assert add_user(self.userdata[0], self.userdata[1], self.userdata[2])
        self.user = get_user(self.userdata[0])
        assert self.user is not None
        assert add_user('UV20@uv20', 'UV20', 'uv20')
        self.user_2 = get_user('UV20@uv20')

    def tearDown(self):
        pass

    def test_create_operation(self):
        with self.app.test_client():
            flight_path = "f3"
            operation = Operation.query.filter_by(path=flight_path).first()
            assert operation is None
            assert self.fm.create_operation(flight_path, "f3 test example", self.user)
            operation = Operation.query.filter_by(path=flight_path).first()
            assert operation.id is not None
            assert operation.path == "f3"

    def test_list_operations(self):
        with self.app.test_client():
            operations = ["alpha", "beta", "gamma"]
            for fp in operations:
                assert self.fm.create_operation(fp, f"{fp} test example", self.user)
            assert len(self.fm.list_operations(self.user)) == 3
            assert len(self.fm.list_operations(self.user_2)) == 0
            fps = self.fm.list_operations(self.user)
            all_operations = [fp['path'] for fp in fps]
            assert operations == all_operations

    def test_get_operation_details(self):
        with self.app.test_client():
            description = "test example"
            flight_path, operation = self._create_operation(flight_path="V1", description=description)
            details = self.fm.get_operation_details(operation.id, self.user)
            assert details["description"] == description
            assert details["path"] == flight_path
            assert details["id"] == operation.id

    def test_get_authorized_users(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V1")
            users = self.fm.get_authorized_users(operation.id)
            assert users[0] == {'username': 'UV10', 'access_level': 'creator'}

    def test_fetch_users_without_permission(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V2")
            assert self.fm.fetch_users_without_permission(operation.id, self.user_2.id) is False
            without_permission = self.fm.fetch_users_without_permission(operation.id, self.user.id)
            # ToDo after seeding removed use absolut comparison
            assert without_permission[-1] == [self.user_2.username, self.user_2.id]

    def test_fetch_users_with_permission(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V3")
            assert self.fm.fetch_users_with_permission(operation.id, self.user_2.id) is False
            # we look in the query only on others than creator
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == []

    def test_add_bulk_permissions(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V4")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == []
            self.fm.add_bulk_permission(operation.id, self.user, [self.user_2.id], "viewer")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == [[self.user_2.username, 'viewer', self.user_2.id]]

    def test_modify_bulk_permissions(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V5")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == []
            self.fm.add_bulk_permission(operation.id, self.user, [self.user_2.id], "viewer")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == [[self.user_2.username, 'viewer', self.user_2.id]]
            self.fm.modify_bulk_permission(operation.id, self.user, [self.user_2.id], "collaborator")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == [[self.user_2.username, 'collaborator', self.user_2.id]]

    def test_delete_bulk_permissions(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V6")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == []
            self.fm.add_bulk_permission(operation.id, self.user, [self.user_2.id], "viewer")
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == [[self.user_2.username, 'viewer', self.user_2.id]]
            assert self.fm.delete_bulk_permission(operation.id, self.user, [self.user_2.id])
            with_permission = self.fm.fetch_users_with_permission(operation.id, self.user.id)
            assert with_permission == []

    def test_import_permissions(self):
        with self.app.test_client():
            flight_path, operation_1 = self._create_operation(flight_path="V7")
            with_permission = self.fm.fetch_users_with_permission(operation_1.id, self.user.id)
            assert with_permission == []
            self.fm.add_bulk_permission(operation_1.id, self.user, [self.user_2.id], "viewer")

            flight_path, operation_2 = self._create_operation(flight_path="V8")
            with_permission = self.fm.fetch_users_with_permission(operation_2.id, self.user.id)
            assert with_permission == []

            self.fm.import_permissions(operation_1.id, operation_2.id, self.user.id)
            with_permission = self.fm.fetch_users_with_permission(operation_2.id, self.user.id)
            assert with_permission == [[self.user_2.username, 'viewer', self.user_2.id]]

    def test_update_operation(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V9")
            new_flight_path = 'NEW_V9'
            assert self.fm.update_operation(operation.id, 'path', new_flight_path, self.user)
            operation = Operation.query.filter_by(path=new_flight_path).first()
            assert operation.path == new_flight_path
            data = fs.open_fs(self.fm.data_dir)
            assert data.exists(new_flight_path)
            new_description = "my new description"
            assert self.fm.update_operation(operation.id, 'description', new_description, self.user)
            operation = Operation.query.filter_by(path=new_flight_path).first()
            assert operation.description == new_description

    def test_delete_file(self):
        # ToDo rename to operation
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V10")
            assert operation.path == flight_path
            assert self.fm.delete_file(operation.id, self.user)
            operation = Operation.query.filter_by(path=flight_path).first()
            assert operation is None

    def test_get_all_changes(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V11")
            assert self.fm.save_file(operation.id, "content1", self.user)
            assert self.fm.save_file(operation.id, "content2", self.user)
            all_changes = self.fm.get_all_changes(operation.id, self.user)
            assert len(all_changes) == 2

    def test_get_change_content(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V12", content='initial')
            assert self.fm.save_file(operation.id, "content1", self.user)
            assert self.fm.save_file(operation.id, "content2", self.user)
            assert self.fm.save_file(operation.id, "content3", self.user)
            all_changes = self.fm.get_all_changes(operation.id, self.user)
            previous_change = self.fm.get_change_content(all_changes[2]["id"])
            assert previous_change == "content1"
            previous_change = self.fm.get_change_content(all_changes[1]["id"])
            assert previous_change == "content2"

    def test_set_version_name(self):
        with self.app.test_client():
            flight_path, operation = self._create_operation(flight_path="V13", content='initial')
            assert self.fm.save_file(operation.id, "content1", self.user)
            all_changes = self.fm.get_all_changes(operation.id, self.user)
            ch_id = all_changes[-1]["id"]
            self.fm.set_version_name(ch_id, operation.id, self.user.id, "berlin")
            all_changes = self.fm.get_all_changes(operation.id, self.user)
            version_name = all_changes[-1]["version_name"]
            assert version_name == "berlin"

    def _create_operation(self, flight_path="firstflight", description="example", user=None, content=None):
        if user is None:
            user = self.user
        self.fm.create_operation(flight_path, description, user, content=content)
        operation = Operation.query.filter_by(path=flight_path).first()
        return flight_path, operation



# -*- coding: utf-8 -*-
"""

    mslib.mswms._tests.test_demodata
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests mswms.demodata

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2017 Reimar Bauer
    :copyright: Copyright 2016-2022 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from past.builtins import basestring

import numpy as np
from mslib._tests.constants import SERVER_CONFIG_FS, DATA_FS, ROOT_FS, SERVER_CONFIG_FILE
import mslib.mswms.demodata as demodata


class TestDemodata(object):
    def test_data_creation(self):
        assert ROOT_FS.exists(u'.')
        assert DATA_FS.exists(u'.')
        assert SERVER_CONFIG_FS.exists(SERVER_CONFIG_FILE)
        assert len(DATA_FS.listdir(u'.')) == 23

    def test_get_profile(self):
        mean, std = demodata.get_profile("air_pressure", [1000, 10000, 50000], "air_temperature")
        assert np.allclose(mean, [223., 212., 255.])
        assert np.allclose(std, [3.03, 6.31, 6.33])

        mean, std = demodata.get_profile("air_potential_temperature", [300, 350, 400], "geopotential_height")
        assert np.allclose(mean, [2970., 12026.92307692, 15307.89473684])
        assert np.allclose(std, [118, 247.69230769, 164.34210526])

        mean, std = demodata.get_profile("ertel_potential_vorticity", [2, 4, 8], "geopotential_height")
        assert np.allclose(mean, [9565.89928058, 11148.1865285, 14139.43661972])
        assert np.allclose(std, [255.705035971, 257.943005181, 196.591549296])

    def test_generate_field(self):
        data, unit = demodata.generate_field("air_pressure", [10, 100, 500], "geopotential_height", 2, 4, 5)
        assert isinstance(data, np.ndarray)
        assert isinstance(unit, basestring)
        assert len(data.shape) == 4
        assert all(_x == _y for _x, _y in zip(data.shape, (2, 3, 4, 5)))

    def test_generate_surface(self):
        data, unit = demodata.generate_surface("atmosphere_boundary_layer_thickness", 2, 4, 5)
        assert isinstance(data, np.ndarray)
        assert isinstance(unit, basestring)
        assert len(data.shape) == 3
        assert all(_x == _y for _x, _y in zip(data.shape, (2, 4, 5)))

    def test_SURFACE(self):
        assert isinstance(demodata._SURFACE, dict)
        for key, entry in list(demodata._SURFACE.items()):
            assert "data" in entry
            assert "unit" in entry
            assert isinstance(entry["unit"], basestring)
            assert isinstance(entry["data"], np.ndarray)

    def test_PROFILES(self):
        assert isinstance(demodata._PROFILES, dict)
        for key, entry in list(demodata._PROFILES.items()):
            assert "data" in entry
            assert "unit" in entry
            assert isinstance(entry["unit"], basestring)
            assert isinstance(entry["data"], np.ndarray)
